# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-arguments
from collections import defaultdict
import io
import os
import random
import struct
from contextlib import contextmanager, suppress
import logging
import socket
from threading import Condition
import time
from typing import (
    BinaryIO, Collection, Dict, FrozenSet, Iterable, Iterator, List, Optional,
    Sequence, Set, Tuple, TypeVar, Union, cast)
from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod)
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.typing.coords import XY
from spinn_machine import (
    CoreSubsets, Machine, MulticastRoutingEntry, RoutingEntry)
from spinn_machine.tags import AbstractTag, IPTag, ReverseIPTag
from spinnman.connections.abstract_classes import Connection
from spinnman.connections.udp_packet_connections import SDPConnection
from spinnman.constants import (
    BMP_POST_POWER_ON_SLEEP_TIME, BMP_POWER_ON_TIMEOUT, BMP_TIMEOUT,
    CPU_MAX_USER, CPU_USER_OFFSET, CPU_USER_START_ADDRESS,
    IPTAG_TIME_OUT_WAIT_TIMES, SCP_SCAMP_PORT, SYSTEM_VARIABLE_BASE_ADDRESS,
    UDP_BOOT_CONNECTION_DEFAULT_PORT, NO_ROUTER_DIAGNOSTIC_FILTERS,
    ROUTER_REGISTER_BASE_ADDRESS, ROUTER_DEFAULT_FILTERS_MAX_POSITION,
    ROUTER_FILTER_CONTROLS_OFFSET, ROUTER_DIAGNOSTIC_FILTER_SIZE, N_RETRIES,
    BOOT_RETRIES, POWER_CYCLE_WAIT_TIME_IN_SECONDS)
from spinnman.data import SpiNNManDataView
from spinnman.exceptions import (
    SpinnmanInvalidParameterException, SpinnmanException, SpinnmanIOException,
    SpinnmanTimeoutException, SpinnmanGenericProcessException,
    SpinnmanUnexpectedResponseCodeException,
    SpiNNManCoresNotInStateException)
from spinnman.model import (
    CPUInfo, CPUInfos, DiagnosticFilter, ChipSummaryInfo,
    IOBuffer, MachineDimensions, RouterDiagnostics, VersionInfo)
from spinnman.model.enums import (
    CPUState, SDP_PORTS, SDP_RUNNING_MESSAGE_CODES, UserRegister)
from spinnman.messages.scp.abstract_messages import AbstractSCPResponse
from spinnman.messages.scp.enums import Signal
from spinnman.messages.scp.impl.get_chip_info import GetChipInfo
from spinnman.messages.scp.impl.get_chip_info_response import (
    GetChipInfoResponse)
from spinnman.messages.sdp import SDPFlag, SDPHeader, SDPMessage
from spinnman.messages.spinnaker_boot import (
    SystemVariableDefinition, SpinnakerBootMessages)
from spinnman.messages.scp.enums import PowerCommand
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.impl import (
    BMPGetVersion, SetPower, ReadFPGARegister,
    WriteFPGARegister, IPTagSetTTO, ReverseIPTagSet,
    WriteMemory, SendSignal, AppStop,
    IPTagSet, IPTagClear, RouterClear, DoSync)
from spinnman.processes import ConnectionSelector
from spinnman.connections.udp_packet_connections import (
    BMPConnection, BootConnection, SCAMPConnection)
from spinnman.processes import (
    GetMachineProcess, GetVersionProcess,
    MallocSDRAMProcess, WriteMemoryProcess, ReadMemoryProcess,
    GetCPUInfoProcess, GetExcludeCPUInfoProcess, GetIncludeCPUInfoProcess,
    ReadIOBufProcess, ApplicationRunProcess,
    LoadFixedRouteRoutingEntryProcess, FixedConnectionSelector,
    ReadFixedRouteRoutingEntryProcess,
    LoadMultiCastRoutesProcess, GetTagsProcess, GetMultiCastRoutesProcess,
    SendSingleCommandProcess, ReadRouterDiagnosticsProcess,
    MostDirectConnectionSelector, ApplicationCopyRunProcess,
    GetNCoresInStateProcess)
from spinnman.transceiver.transceiver import Transceiver
from spinnman.transceiver.extendable_transceiver import ExtendableTransceiver
from spinnman.utilities.utility_functions import get_vcpu_address

#: Type of a response.
# This allows subclasses to be used
_AbstractSCPResponse = TypeVar(
    "_AbstractSCPResponse", bound=AbstractSCPResponse)

logger = FormatAdapter(logging.getLogger(__name__))

_SCAMP_NAME = "SC&MP"
_SCAMP_VERSION = (4, 0, 0)

_BMP_NAME = "BC&MP"
_BMP_MAJOR_VERSIONS = [1, 2]

_CONNECTION_CHECK_RETRIES = 3
INITIAL_FIND_SCAMP_RETRIES_COUNT = 3

_TWO_BYTES = struct.Struct("<BB")
_FOUR_BYTES = struct.Struct("<BBBB")
_ONE_WORD = struct.Struct("<I")
_ONE_LONG = struct.Struct("<Q")
_EXECUTABLE_ADDRESS = 0x67800000

_POWER_CYCLE_WARNING = (
    "When power-cycling a board, it is recommended that you wait for 30 "
    "seconds before attempting a reboot. Therefore, the tools will now "
    "wait for 30 seconds. If you wish to avoid this wait, please set "
    "reset_machine_on_startup = False in the [Machine] section of the "
    "relevant configuration (cfg) file.")

_POWER_CYCLE_FAILURE_WARNING = (
    "The end user requested the power-cycling of the board. But the "
    "tools did not have the required BMP connection to facilitate a "
    "power-cycling, and therefore will not do so. please set the "
    "bmp_names accordingly in the [Machine] section of the relevant "
    "configuration (cfg) file. Or use a machine assess process which "
    "provides the BMP data (such as a spalloc system) or finally set "
    "reset_machine_on_startup = False in the [Machine] section of the "
    "relevant configuration (cfg) file to avoid this warning in future.")


class BaseTransceiver(ExtendableTransceiver, metaclass=AbstractBase):
    """
    A base for all the code shared by all Version of the Transceiver.

    """
    __slots__ = (
        "_all_connections",
        "_bmp_selector",
        "_bmp_connection",
        "_boot_send_connection",
        "_chip_execute_lock_condition",
        "_chip_execute_locks",
        "_height",
        "_iobuf_size",
        "_machine_off",
        "_n_chip_execute_locks",
        "_scamp_connection_selector",
        "_scamp_connections",
        "_udp_scamp_connections",
        "_width")

    def __init__(self, connections: Optional[List[Connection]] = None,
                 power_cycle: bool = False):
        """
        :param list(Connection) connections:
            An iterable of connections to the board.  If not specified, no
            communication will be possible until connections are found.
        :param bool power_cycle: If True will power cycle the machine:
        :raise SpinnmanIOException:
            If there is an error communicating with the board, or if no
            connections to the board can be found (if connections is ``None``)
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        super().__init__()

        # Place to keep the current machine
        self._width: Optional[int] = None
        self._height: Optional[int] = None
        self._iobuf_size: Optional[int] = None

        # A set of the original connections - used to determine what can
        # be closed
        if connections is None:
            connections = list()

        # A set of all connection - used for closing
        self._all_connections = set(connections)

        # A boot send connection - there can only be one in the current system,
        # or otherwise bad things can happen!
        self._boot_send_connection: Optional[BootConnection] = None

        # A dict of IP address -> SCAMP connection
        # These are those that can be used for setting up IP Tags
        self._udp_scamp_connections: Dict[str, SCAMPConnection] = dict()

        # A list of all connections that can be used to send and receive SCP
        # messages for SCAMP interaction
        self._scamp_connections: List[SCAMPConnection] = list()

        # The BMP connections
        self._bmp_connection: Optional[BMPConnection] = None

        # A lock against single chip executions (entry is (x, y))
        # The condition should be acquired before the locks are
        # checked or updated
        # The write lock condition should also be acquired to avoid a flood
        # fill during an individual chip execute
        self._chip_execute_locks: Dict[
            Tuple[int, int], Condition] = defaultdict(Condition)
        self._chip_execute_lock_condition = Condition()
        self._n_chip_execute_locks = 0

        # build connection selectors for the processes.
        self._bmp_selector: Optional[
            FixedConnectionSelector[BMPConnection]] = None
        self._scamp_connection_selector = \
            self.__identify_connections(connections)

        # Check that the BMP connections are valid
        self.__check_bmp_connection()

        self._machine_off = False
        if power_cycle:
            self._power_off_machine()
        self._ensure_board_is_ready()

    @property
    @overrides(ExtendableTransceiver.bmp_selector)
    def bmp_selector(self) -> Optional[FixedConnectionSelector[BMPConnection]]:
        return self._bmp_selector

    @property
    @overrides(ExtendableTransceiver.scamp_connection_selector)
    def scamp_connection_selector(self) -> MostDirectConnectionSelector:
        return self._scamp_connection_selector

    def _where_is_xy(self, x: int, y: int) -> Optional[str]:
        """
        Attempts to get where_is_x_y info from the machine

        If no machine will do its best.

        :param int x:
        :param int y:
        :rtype: str
        """
        try:
            if SpiNNManDataView.has_machine():
                return SpiNNManDataView.get_machine().where_is_xy(x, y)
            return (f"No Machine. "
                    f"Root IP:{self._scamp_connections[0].remote_ip_address}"
                    f"x:{x} y:{y}")
        except Exception as ex:  # pylint: disable=broad-except
            return str(ex)

    def __identify_connections(
            self, connections: Iterable[Connection]
            ) -> MostDirectConnectionSelector:
        for conn in connections:

            # locate the only boot send connection
            if isinstance(conn, BootConnection):
                if self._boot_send_connection is not None:
                    raise SpinnmanInvalidParameterException(
                        "connections", f"[... {conn} ...]",
                        "Only a single SpinnakerBootSender can be specified")
                self._boot_send_connection = conn

            # Locate any connections that talk to a BMP
            if isinstance(conn, BMPConnection):
                # If it is a BMP connection, add it here
                if self._bmp_connection is not None:
                    raise NotImplementedError(
                        "Only one BMP connection supported")
                self._bmp_connection = conn
                self._bmp_selector = FixedConnectionSelector(conn)
            # Otherwise, check if it can send and receive SCP (talk to SCAMP)
            elif isinstance(conn, SCAMPConnection):
                self._scamp_connections.append(conn)

        # update the transceiver with the connection selectors.
        return MostDirectConnectionSelector(self._scamp_connections)

    def __check_bmp_connection(self) -> None:
        """
        Check that the BMP connections are actually connected to valid BMPs.

        :raise SpinnmanIOException: when a connection is not linked to a BMP
        """
        # check that the UDP BMP connection is actually connected to a BMP
        # via the get_scamp_version command
        if self._bmp_connection is not None:
            conn = self._bmp_connection

            try:
                version_info = self._get_scamp_version(
                    conn.chip_x, conn.chip_y, self._bmp_selector)
                fail_version_name = version_info.name != _BMP_NAME
                fail_version_num = \
                    version_info.version_number[0] not in _BMP_MAJOR_VERSIONS
                if fail_version_name or fail_version_num:
                    raise SpinnmanIOException(
                        f"The BMP at {conn.remote_ip_address} is running "
                        f"{version_info.name} {version_info.version_string} "
                        "which is incompatible with this transceiver, required"
                        f" version is {_BMP_NAME} {_BMP_MAJOR_VERSIONS}")

                logger.info("Using BMP at {} with version {} {}",
                            conn.remote_ip_address, version_info.name,
                            version_info.version_string)

            # If it fails to respond due to timeout, maybe that the connection
            # isn't valid
            except SpinnmanTimeoutException as e:
                raise SpinnmanException(
                    f"BMP connection to {conn.remote_ip_address} is "
                    "not responding") from e
            except Exception:
                logger.exception("Failed to speak to BMP at {}",
                                 conn.remote_ip_address)
                raise

    @staticmethod
    def _check_connection(
            connection: SCAMPConnection) -> Optional[ChipSummaryInfo]:
        """
        Check that the given connection to the given chip works.

        :param ConnectionSelector connection_selector:
            the connection selector to use
        :param int chip_x: the chip x coordinate to try to talk to
        :param int chip_y: the chip y coordinate to try to talk to
        :return: True if a valid response is received, False otherwise
        :rtype: ChipInfo or None
        """
        chip_x, chip_y = connection.chip_x, connection.chip_y
        connection_selector = FixedConnectionSelector(connection)
        for _ in range(_CONNECTION_CHECK_RETRIES):
            try:
                sender: SendSingleCommandProcess[GetChipInfoResponse] = \
                    SendSingleCommandProcess(connection_selector)
                chip_info = sender.execute(  # pylint: disable=no-member
                    GetChipInfo(chip_x, chip_y)).chip_info
                if not chip_info.is_ethernet_available:
                    time.sleep(0.1)
                else:
                    return chip_info
            except (SpinnmanGenericProcessException, SpinnmanTimeoutException,
                    SpinnmanUnexpectedResponseCodeException):
                pass
            except SpinnmanIOException:
                break
        return None

    @overrides(Transceiver.send_sdp_message)
    def send_sdp_message(self, message: SDPMessage,
                         connection: Optional[SDPConnection] = None):
        if connection is None:
            connection_to_use: SDPConnection = self._scamp_connections[
                random.randint(0, len(self._scamp_connections) - 1)]
        else:
            connection_to_use = connection
        connection_to_use.send_sdp_message(message)

    def _check_and_add_scamp_connections(
            self, x: int, y: int, ip_address: str):
        """
        :param int x: X coordinate of target chip
        :param int y: Y coordinate of target chip
        :param str ip_address: IP address of target chip

        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        conn = SCAMPConnection(remote_host=ip_address, chip_x=x, chip_y=y)

        # check if it works
        chip_info = self._check_connection(conn)
        if chip_info is not None and chip_info.ethernet_ip_address is not None:
            self._all_connections.add(conn)
            self._udp_scamp_connections[chip_info.ethernet_ip_address] = conn
            self._scamp_connections.append(conn)
        else:
            logger.warning(
                "Additional Ethernet connection on {} at chip {}, {} "
                "cannot be contacted", ip_address, x, y)

    @overrides(Transceiver.discover_scamp_connections)
    def discover_scamp_connections(self) -> None:
        # Get the machine dimensions
        dims = self._get_machine_dimensions()

        # Find all the new connections via the machine Ethernet-connected chips
        version = SpiNNManDataView.get_machine_version()
        for x, y in version.get_potential_ethernet_chips(
                dims.width, dims.height):
            ip_addr_item = SystemVariableDefinition.ethernet_ip_address
            try:
                # TODO avoid here_is_x,y if read_memory fails
                data = self.read_memory(
                    x, y,
                    SYSTEM_VARIABLE_BASE_ADDRESS + ip_addr_item.offset, 4)
            except SpinnmanGenericProcessException:
                continue
            ip = _FOUR_BYTES.unpack_from(data)
            ip_address = f"{ip[0]}.{ip[1]}.{ip[2]}.{ip[3]}"
            logger.info(ip_address)
            self._check_and_add_scamp_connections(x, y, ip_address)
        self._scamp_connection_selector = MostDirectConnectionSelector(
            self._scamp_connections)

    @overrides(Transceiver.add_scamp_connections)
    def add_scamp_connections(self, connections: Dict[XY, str]):
        for ((x, y), ip_address) in connections.items():
            self._check_and_add_scamp_connections(x, y, ip_address)
        self._scamp_connection_selector = MostDirectConnectionSelector(
            self._scamp_connections)

    @overrides(Transceiver.get_connections)
    def get_connections(self) -> Set[Connection]:
        return self._all_connections

    def _get_machine_dimensions(self) -> MachineDimensions:
        """
        Get the maximum chip X-coordinate and maximum chip Y-coordinate of
        the chips in the machine.

        :return: The dimensions of the machine
        :rtype: MachineDimensions
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        if self._width is None or self._height is None:
            height_item = SystemVariableDefinition.y_size
            self._height, self._width = _TWO_BYTES.unpack_from(
                self.read_memory(
                    AbstractSCPRequest.DEFAULT_DEST_X_COORD,
                    AbstractSCPRequest.DEFAULT_DEST_Y_COORD,
                    SYSTEM_VARIABLE_BASE_ADDRESS + height_item.offset,
                    2))
        assert self._width is not None and self._height is not None
        return MachineDimensions(self._width, self._height)

    @overrides(Transceiver.get_machine_details)
    def get_machine_details(self) -> Machine:
        # Get the width and height of the machine
        dims = self._get_machine_dimensions()

        # Get the coordinates of the boot chip
        version_info = self._get_scamp_version()

        # Get the details of all the chips
        get_machine_process = GetMachineProcess(
            self._scamp_connection_selector)
        machine = get_machine_process.get_machine_details(
            version_info.x, version_info.y, dims.width, dims.height)

        # Work out and add the SpiNNaker links and FPGA links
        machine.add_spinnaker_links()
        machine.add_fpga_links()

        if self._boot_send_connection:
            logger.info(f"Detected {machine.summary_string()}")
        return machine

    def _get_scamp_version(
            self, chip_x: int = AbstractSCPRequest.DEFAULT_DEST_X_COORD,
            chip_y: int = AbstractSCPRequest.DEFAULT_DEST_Y_COORD,
            connection_selector: Optional[ConnectionSelector] = None,
            n_retries: int = N_RETRIES) -> VersionInfo:
        """
        Get the version of SCAMP which is running on the board.

        :param int chip_x: the chip's x coordinate to query for SCAMP version
        :param int chip_y: the chip's y coordinate to query for SCAMP version
        :param ConnectionSelector connection_selector:
            the connection to send the SCAMP version
            or `None` (if `None` then a random SCAMP connection is used).
        :param int n_retries:
        :return: The version identifier
        :rtype: VersionInfo
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidParameterException:
            If the timeout is less than 1
        :raise SpinnmanTimeoutException:
            If none of the retries resulted in a response before the timeout
            (suggesting that the board is not booted).
        """
        if connection_selector is None:
            connection_selector = self._scamp_connection_selector
        process = GetVersionProcess(connection_selector, n_retries)
        return process.get_version(x=chip_x, y=chip_y, p=0)

    @property
    @abstractmethod
    def boot_led_0_value(self) -> int:
        """
        The Values to be set in SpinnakerBootMessages for led_0

        :rtype int:
        """
        raise NotImplementedError

    def _boot_board(self, extra_boot_values: Optional[Dict[
            SystemVariableDefinition, object]] = None):
        """
        Attempt to boot the board. No check is performed to see if the
        board is already booted.

        :param dict(SystemVariableDefinition,object) extra_boot_values:
            extra values to set during boot
            Any additional or overwrite values to set during boot.
            This should only be used for values which are not standard
            based on the board version.
        :raise SpinnmanInvalidParameterException:
            If the board version is not known
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        """
        if not self._boot_send_connection:
            # No can do. Can't boot without a boot connection.
            raise SpinnmanIOException("no boot connection available")
        if extra_boot_values is None:
            extra_boot_values = dict()
        if SystemVariableDefinition.led_0 not in extra_boot_values:
            extra_boot_values[SystemVariableDefinition.led_0] = \
                self.boot_led_0_value
        boot_messages = SpinnakerBootMessages(
            extra_boot_values=extra_boot_values)
        for boot_message in boot_messages.messages:
            self._boot_send_connection.send_boot_message(boot_message)
        time.sleep(2.0)

    def _call(self, req: AbstractSCPRequest[_AbstractSCPResponse],
              **kwargs) -> _AbstractSCPResponse:
        """
        Wrapper that makes doing simple SCP calls easier,
        especially with types.
        """
        proc: SendSingleCommandProcess[_AbstractSCPResponse] = \
            SendSingleCommandProcess(self._scamp_connection_selector, **kwargs)
        return proc.execute(req)

    @staticmethod
    def _is_scamp_version_compabible(version: Tuple[int, int, int]) -> bool:
        """
        Determine if the version of SCAMP is compatible with this transceiver.

        :param tuple(int,int,int) version: The version to test
        :rtype: bool
        """
        # The major version must match exactly
        if version[0] != _SCAMP_VERSION[0]:
            return False

        # If the minor version matches, the patch version must be >= the
        # required version
        if version[1] == _SCAMP_VERSION[1]:
            return version[2] >= _SCAMP_VERSION[2]

        # If the minor version is > than the required version, the patch
        # version is irrelevant
        return version[1] > _SCAMP_VERSION[1]

    def _ensure_board_is_ready(
            self, n_retries: int = 5, extra_boot_values: Optional[Dict[
            SystemVariableDefinition, object]] = None):
        """
        Ensure that the board is ready to interact with this version of the
        transceiver. Boots the board if not already booted and verifies that
        the version of SCAMP running is compatible with this transceiver.

        :param int n_retries: The number of times to retry booting
        :param dict(SystemVariableDefinition,object) extra_boot_values:
            Any additional or overwrite values to set during boot.
            This should only be used for values which are not standard
            based on the board version.
        :raise SpinnmanIOException:
            * If there is a problem booting the board
            * If the version of software on the board is not compatible with
              this transceiver
        """
        logger.info("Working out if machine is booted")
        if self._machine_off:
            version_info = None
        else:
            version_info = self._try_to_find_scamp_and_boot(
                INITIAL_FIND_SCAMP_RETRIES_COUNT, extra_boot_values)

        # If we fail to get a SCAMP version this time, try other things
        if version_info is None and self._bmp_connection is not None:
            # start by powering up each BMP connection
            logger.info("Attempting to power on machine")
            self._power_on_machine()

            # Sleep a bit to let things get going
            time.sleep(2.0)
            logger.info("Attempting to boot machine")

            # retry to get a SCAMP version, this time trying multiple times
            version_info = self._try_to_find_scamp_and_boot(
                n_retries, extra_boot_values)

        # verify that the version is the expected one for this transceiver
        if version_info is None:
            raise SpinnmanIOException("Failed to communicate with the machine")
        if (version_info.name != _SCAMP_NAME or
                not self._is_scamp_version_compabible(
                    version_info.version_number)):
            raise SpinnmanIOException(
                f"The machine is currently booted with {version_info.name}"
                f" {version_info.version_number} which is incompatible with "
                "this transceiver, required version is "
                f"{_SCAMP_NAME} {_SCAMP_VERSION}")

        logger.info("Machine communication successful")

        # Change the default SCP timeout on the machine, keeping the old one to
        # revert at close
        for scamp_connection in self._scamp_connections:
            self._call(IPTagSetTTO(
                scamp_connection.chip_x, scamp_connection.chip_y,
                IPTAG_TIME_OUT_WAIT_TIMES.TIMEOUT_2560_ms))

            chip_info = self._check_connection(scamp_connection)
            if chip_info is not None and chip_info.ethernet_ip_address:
                self._udp_scamp_connections[chip_info.ethernet_ip_address] = \
                    scamp_connection

        # Update the connection selector so that it can ask for processor ids
        self._scamp_connection_selector = MostDirectConnectionSelector(
            self._scamp_connections)

    def __is_default_destination(self, version_info: VersionInfo) -> bool:
        return (version_info.x == AbstractSCPRequest.DEFAULT_DEST_X_COORD
                and version_info.y == AbstractSCPRequest.DEFAULT_DEST_Y_COORD)

    def _try_to_find_scamp_and_boot(
            self, tries_to_go: int, extra_boot_values: Optional[Dict[
                SystemVariableDefinition, object]]) -> Optional[VersionInfo]:
        """
        Try to detect if SCAMP is running, and if not, boot the machine.

        :param int tries_to_go: how many attempts should be supported
        :param dict(SystemVariableDefinition,object) extra_boot_values:
            Any additional or overwrite values to set during boot.
            This should only be used for values which are not standard
            based on the board version.
        :return: version info
        :rtype: VersionInfo
        :raise SpinnmanIOException:
            If there is a problem communicating with the machine
        """
        version_info = None
        current_tries_to_go = tries_to_go
        while version_info is None and current_tries_to_go > 0:
            try:
                version_info = self._get_scamp_version(n_retries=BOOT_RETRIES)
                if self.__is_default_destination(version_info):
                    version_info = None
                    time.sleep(0.1)
            except SpinnmanGenericProcessException as e:
                if isinstance(e.exception, SpinnmanTimeoutException):
                    logger.info("Attempting to boot machine")
                    self._boot_board(extra_boot_values)
                    current_tries_to_go -= 1
                elif isinstance(e.exception, SpinnmanIOException):
                    raise SpinnmanIOException(
                        "Failed to communicate with the machine") from e
                else:
                    raise
            except SpinnmanTimeoutException:
                logger.info("Attempting to boot machine")
                self._boot_board(extra_boot_values)
                current_tries_to_go -= 1
            except SpinnmanIOException as e:
                raise SpinnmanIOException(
                    "Failed to communicate with the machine") from e

        # The last thing we tried was booting, so try again to get the version
        if version_info is None:
            with suppress(SpinnmanException):
                version_info = self._get_scamp_version()
                if self.__is_default_destination(version_info):
                    version_info = None
        if version_info is not None:
            logger.info("Found board with version {}", version_info)
        return version_info

    @overrides(Transceiver.get_cpu_infos)
    def get_cpu_infos(
            self, core_subsets: Optional[CoreSubsets] = None,
            states: Union[CPUState, Iterable[CPUState], None] = None,
            include: bool = True) -> CPUInfos:
        # Get all the cores if the subsets are not given
        if core_subsets is None:
            core_subsets = CoreSubsets()
            for chip in SpiNNManDataView.get_machine().chips:
                for p in chip.all_processor_ids:
                    core_subsets.add_processor(chip.x, chip.y, p)

        if states is None:
            process = GetCPUInfoProcess(self._scamp_connection_selector)
        else:
            if isinstance(states, CPUState):
                state_set = frozenset((states, ))
            else:
                state_set = frozenset(states)
            if include:
                process = GetIncludeCPUInfoProcess(
                    self._scamp_connection_selector, state_set)
            else:
                process = GetExcludeCPUInfoProcess(
                    self._scamp_connection_selector, state_set)

        cpu_info = process.get_cpu_info(core_subsets)
        return cpu_info

    @overrides(Transceiver.get_clock_drift)
    def get_clock_drift(self, x: int, y: int) -> float:
        drift_fp = 1 << 17

        drift_b = self._get_sv_data(x, y, SystemVariableDefinition.clock_drift)
        drift = struct.unpack("<i", struct.pack("<I", drift_b))[0]
        return drift / drift_fp

    def _get_sv_data(
            self, x: int, y: int,
            data_item: SystemVariableDefinition) -> Union[int, bytes]:
        """
        :param int x:
        :param int y:
        :param SystemVariableDefinition data_item:
        """
        addr = SYSTEM_VARIABLE_BASE_ADDRESS + data_item.offset
        if data_item.data_type.is_byte_array:
            size = cast(int, data_item.array_size)
            # Do not need to decode the bytes of a byte array
            return self.read_memory(x, y, addr, size)
        return struct.unpack_from(
            data_item.data_type.struct_code,
            self.read_memory(x, y, addr, data_item.data_type.value))[0]

    @staticmethod
    def __get_user_register_address_from_core(p: int, user: UserRegister):
        """
        Get the address of user *N* for a given processor on the board.

        .. note::
            Conventionally, user_0 usually holds the address of the table of
            memory regions.

        :param int p: The ID of the processor to get the user N address from
        :param int user: The user "register" number to get the address for
        :return: The address for user N register for this processor
        :rtype: int
        """
        if user < 0 or user > CPU_MAX_USER:
            raise ValueError(
                f"Incorrect user number {user}")
        return (get_vcpu_address(p) + CPU_USER_START_ADDRESS +
                CPU_USER_OFFSET * user)

    @overrides(Transceiver.read_user)
    def read_user(self, x: int, y: int, p: int, user: UserRegister):
        addr = self.__get_user_register_address_from_core(p, user)
        return self.read_word(x, y, addr)

    @overrides(Transceiver.add_cpu_information_from_core)
    def add_cpu_information_from_core(
            self, cpu_infos: CPUInfos, x: int, y: int, p: int,
            states: Iterable[CPUState]):
        core_subsets = CoreSubsets()
        core_subsets.add_processor(x, y, p)
        new_infos = self.get_cpu_infos(core_subsets)
        cpu_infos.add_infos(new_infos, states)

    @overrides(Transceiver.get_region_base_address)
    def get_region_base_address(self, x: int, y: int, p: int):
        return self.read_user(x, y, p, UserRegister.USER_0)

    @overrides(Transceiver.get_iobuf)
    def get_iobuf(self, core_subsets: Optional[CoreSubsets] = None
                  ) -> Iterable[IOBuffer]:
        # making the assumption that all chips have the same iobuf size.
        if self._iobuf_size is None:
            self._iobuf_size = cast(int, self._get_sv_data(
                AbstractSCPRequest.DEFAULT_DEST_X_COORD,
                AbstractSCPRequest.DEFAULT_DEST_Y_COORD,
                SystemVariableDefinition.iobuf_size))
        # Get all the cores if the subsets are not given
        # TODO is core_subsets ever None
        if core_subsets is None:
            core_subsets = CoreSubsets()
            for chip in SpiNNManDataView.get_machine().chips:
                for p in chip.all_processor_ids:
                    core_subsets.add_processor(chip.x, chip.y, p)

        # read iobuf from machine
        process = ReadIOBufProcess(self._scamp_connection_selector)
        return process.read_iobuf(self._iobuf_size, core_subsets)

    @overrides(Transceiver.get_core_state_count)
    def get_core_state_count(
            self, app_id: int, state: CPUState,
            xys: Optional[Iterable[Tuple[int, int]]] = None) -> int:
        process = GetNCoresInStateProcess(self._scamp_connection_selector)
        chip_xys: Iterable[Tuple[int, int]]
        if xys is None:
            machine = SpiNNManDataView.get_machine()
            chip_xys = machine.ethernet_connected_chips
        else:
            chip_xys = xys
        return process.get_n_cores_in_state(chip_xys, app_id, state)

    @contextmanager
    def __flood_execute_lock(self) -> Iterator[Condition]:
        """
        Get a lock for executing a flood fill of an executable.
        """
        # Get the execute lock all together, so nothing can access it
        with self._chip_execute_lock_condition:
            # Wait until nothing is executing
            self._chip_execute_lock_condition.wait_for(
                lambda: self._n_chip_execute_locks < 1)
            yield self._chip_execute_lock_condition

    @overrides(Transceiver.execute_flood)
    def execute_flood(
            self, core_subsets: CoreSubsets,
            executable: Union[BinaryIO, bytes, str], app_id: int, *,
            n_bytes: Optional[int] = None, wait: bool = False,):
        if isinstance(executable, int):
            # No executable is 4 bytes long
            raise TypeError("executable may not be int")
        # Lock against other executables
        with self.__flood_execute_lock():
            # Flood fill the system with the binary
            n_bytes, chksum = self.write_memory(
                0, 0, _EXECUTABLE_ADDRESS, executable, n_bytes=n_bytes,
                get_sum=True)

            # Execute the binary on the cores on 0, 0 if required
            if core_subsets.is_chip(0, 0):
                boot_subset = CoreSubsets()
                boot_subset.add_core_subset(
                    core_subsets.get_core_subset_for_chip(0, 0))
                runner = ApplicationRunProcess(
                    self._scamp_connection_selector)
                runner.run(app_id, boot_subset, wait)

            copy_run = ApplicationCopyRunProcess(
                self._scamp_connection_selector)
            copy_run.run(n_bytes, app_id, core_subsets, chksum, wait)

    def _power_on_machine(self) -> None:
        """
        Power on the whole machine.

        :rtype bool
        :return success of failure to power on machine
        """
        self._power(PowerCommand.POWER_ON)
        # Sleep for 5 seconds as the machine has just been powered on
        time.sleep(BMP_POST_POWER_ON_SLEEP_TIME)

    def _power_off_machine(self) -> None:
        """
        Power off the whole machine.

        :rtype bool
        :return success or failure to power off the machine
        """
        self._power(PowerCommand.POWER_OFF)
        logger.warning(_POWER_CYCLE_WARNING)
        time.sleep(POWER_CYCLE_WAIT_TIME_IN_SECONDS)
        logger.warning("Power cycle wait complete")

    def _bmp_call(self, req: AbstractSCPRequest[_AbstractSCPResponse],
                  **kwargs) -> _AbstractSCPResponse:
        """
        Wrapper that makes doing simple BMP calls easier,
        especially with types.
        """
        if self._bmp_selector is None:
            raise SpinnmanException(
                "this transceiver does not support BMP operations")
        proc: SendSingleCommandProcess[_AbstractSCPResponse] =\
            SendSingleCommandProcess(self._bmp_selector, **kwargs)
        return proc.execute(req)

    def _power(self, power_command: PowerCommand):
        """
        Send a power request to the machine.

        :param PowerCommand power_command: The power command to send
        """
        if self._bmp_connection is None:
            raise NotImplementedError("can not power change without BMP")
        timeout = (
            BMP_POWER_ON_TIMEOUT
            if power_command == PowerCommand.POWER_ON
            else BMP_TIMEOUT)
        self._bmp_call(SetPower(power_command, self._bmp_connection.boards),
                       timeout=timeout, n_retries=0)
        self._machine_off = power_command == PowerCommand.POWER_OFF

    @overrides(Transceiver.read_fpga_register)
    def read_fpga_register(
            self, fpga_num: int, register: int, board: int = 0) -> int:
        response = self._bmp_call(
            ReadFPGARegister(fpga_num, register, board),
            timeout=1.0)
        return response.fpga_register

    @overrides(Transceiver.write_fpga_register)
    def write_fpga_register(
            self, fpga_num: int, register: int, value: int, board: int = 0):
        self._bmp_call(
            WriteFPGARegister(fpga_num, register, value, board))

    @overrides(Transceiver.read_bmp_version)
    def read_bmp_version(self, board: int) -> VersionInfo:
        response = self._bmp_call(BMPGetVersion(board))
        return response.version_info

    @overrides(Transceiver.write_memory)
    def write_memory(
            self, x: int, y: int, base_address: int,
            data: Union[BinaryIO, bytes, int, str], *,
            n_bytes: Optional[int] = None, offset: int = 0, cpu: int = 0,
            get_sum: bool = False) -> Tuple[int, int]:
        process = WriteMemoryProcess(self._scamp_connection_selector)
        if isinstance(data, io.RawIOBase):
            assert n_bytes is not None
            chksum = process.write_memory_from_reader(
                (x, y, cpu), base_address, cast(BinaryIO, data), n_bytes,
                get_sum)
        elif isinstance(data, str):
            if n_bytes is None:
                n_bytes = os.stat(data).st_size
            with open(data, "rb") as reader:
                chksum = process.write_memory_from_reader(
                    (x, y, cpu), base_address, reader, n_bytes, get_sum)
        elif isinstance(data, int):
            n_bytes = 4
            data_to_write = _ONE_WORD.pack(data)
            chksum = process.write_memory_from_bytearray(
                (x, y, cpu), base_address, data_to_write, 0, n_bytes, get_sum)
        else:
            assert isinstance(data, (bytes, bytearray))
            if n_bytes is None:
                n_bytes = len(data)
            chksum = process.write_memory_from_bytearray(
                (x, y, cpu), base_address, data, offset, n_bytes, get_sum)
        return n_bytes, chksum

    @overrides(Transceiver.write_user)
    def write_user(
            self, x: int, y: int, p: int, user: UserRegister, value: int):
        addr = self.__get_user_register_address_from_core(p, user)
        self.write_memory(x, y, addr, int(value))

    @overrides(Transceiver.read_memory)
    def read_memory(
            self, x: int, y: int, base_address: int, length: int,
            cpu: int = 0) -> bytearray:
        try:
            process = ReadMemoryProcess(self._scamp_connection_selector)
            return process.read_memory((x, y, cpu), base_address, length)
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.read_word)
    def read_word(
            self, x: int, y: int, base_address: int, cpu: int = 0) -> int:
        try:
            process = ReadMemoryProcess(self._scamp_connection_selector)
            data = process.read_memory(
                (x, y, cpu), base_address, _ONE_WORD.size)
            (value, ) = _ONE_WORD.unpack(data)
            return value
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.stop_application)
    def stop_application(self, app_id: int):
        if not self._machine_off:
            self._call(AppStop(app_id))
        else:
            logger.warning(
                "You are calling a app stop on a turned off machine. "
                "Please fix and try again")

    def __log_where_is_info(self, cpu_infos: Iterable[
            Union[CPUInfo, Sequence[int]]]):
        """
        Logs the where_is info for each chip in cpu_infos.

        :param cpu_infos:
        """
        xys = set()
        for cpu_info in cpu_infos:
            # TODO: Is it ever not a CPUInfo
            if isinstance(cpu_info, CPUInfo):
                xys.add((cpu_info.x, cpu_info.y))
            else:
                xys.add((cpu_info[0], cpu_info[1]))
        for (x, y) in xys:
            logger.info(self._where_is_xy(x, y))

    @staticmethod
    def __state_set(cpu_states: Union[CPUState, Iterable[CPUState]])\
            -> FrozenSet[CPUState]:
        if isinstance(cpu_states, CPUState):
            return frozenset((cpu_states,))
        else:
            return frozenset(cpu_states)

    @overrides(Transceiver.wait_for_cores_to_be_in_state)
    def wait_for_cores_to_be_in_state(
            self, all_core_subsets: CoreSubsets, app_id: int,
            cpu_states: Union[CPUState, Iterable[CPUState]], *,
            timeout: Optional[float] = None,
            time_between_polls: float = 0.1,
            error_states: FrozenSet[CPUState] = frozenset((
                CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG)),
            counts_between_full_check: int = 100,
            progress_bar: Optional[ProgressBar] = None):
        processors_ready = 0
        max_processors_ready = 0
        timeout_time = None if timeout is None else time.time() + timeout
        tries = 0
        target_states = self.__state_set(cpu_states)
        while (processors_ready < len(all_core_subsets) and
               (timeout_time is None or time.time() < timeout_time)):

            # Get the number of processors in the ready states
            processors_ready = 0
            for cpu_state in target_states:
                processors_ready += self.get_core_state_count(
                    app_id, cpu_state)
            if progress_bar:
                if processors_ready > max_processors_ready:
                    progress_bar.update(
                        processors_ready - max_processors_ready)
                    max_processors_ready = processors_ready
            # If the count is too small, check for error states
            if processors_ready < len(all_core_subsets):
                is_error = False
                for cpu_state in error_states:
                    error_cores = self.get_core_state_count(app_id, cpu_state)
                    if error_cores > 0:
                        is_error = True
                if is_error:
                    error_core_states = self.get_cpu_infos(
                        all_core_subsets, error_states, True)
                    if len(error_core_states) > 0:
                        self.__log_where_is_info(error_core_states)
                        raise SpiNNManCoresNotInStateException(
                            timeout, target_states, error_core_states)

                # If we haven't seen an error, increase the tries, and
                # do a full check if required
                tries += 1
                if tries >= counts_between_full_check:
                    cores_in_state = self.get_cpu_infos(
                        all_core_subsets, target_states, include=True)
                    # convert to a list of xyp values
                    cores_in_state_xyps = list(cores_in_state)
                    processors_ready = len(cores_in_state_xyps)
                    tries = 0

                    # iterate over the cores waiting to finish and see
                    # which ones we're missing
                    if get_config_bool("Machine", "report_waiting_logs"):
                        for core_subset in all_core_subsets.core_subsets:
                            for p in core_subset.processor_ids:
                                if ((core_subset.x, core_subset.y, p) not in
                                        cores_in_state_xyps):
                                    logger.warning(
                                        "waiting on {}:{}:{}",
                                        core_subset.x, core_subset.y, p)

                # If we're still not in the correct state, wait a bit
                if processors_ready < len(all_core_subsets):
                    time.sleep(time_between_polls)

        # If we haven't reached the final state, do a final full check
        if processors_ready < len(all_core_subsets):
            cores_not_in_state = self.get_cpu_infos(
                all_core_subsets, cpu_states, include=False)

            # If we are sure we haven't reached the final state,
            # report a timeout error
            if len(cores_not_in_state) != 0:
                self.__log_where_is_info(cores_not_in_state)
                raise SpiNNManCoresNotInStateException(
                    timeout, target_states, cores_not_in_state)

    @overrides(Transceiver.send_signal)
    def send_signal(self, app_id: int, signal: Signal):
        self._call(SendSignal(app_id, signal))

    def _locate_spinnaker_connection_for_board_address(
            self, board_address: str) -> Optional[SCAMPConnection]:
        """
        Find a connection that matches the given board IP address.

        :param str board_address:
            The IP address of the Ethernet connection on the board
        :return: A connection for the given IP address, or `None` if no such
            connection exists
        :rtype: SCAMPConnection
        """
        return self._udp_scamp_connections.get(board_address, None)

    @overrides(Transceiver.set_ip_tag)
    def set_ip_tag(self, ip_tag: IPTag, use_sender: bool = False):
        # Check that the tag has a port assigned
        if ip_tag.port is None:
            raise SpinnmanInvalidParameterException(
                "ip_tag.port", "None", "The tag port must have been set")

        # Get the connections - if the tag specifies a connection, use that,
        # otherwise apply the tag to all connections
        connections = self.__get_connection_list(None, ip_tag.board_address)
        if not connections:
            raise SpinnmanInvalidParameterException(
                "ip_tag", str(ip_tag),
                "The given board address is not recognised")

        for connection in connections:
            # Convert the host string
            host_string = ip_tag.ip_address
            if host_string in ("localhost", ".", "0.0.0.0"):
                host_string = connection.local_ip_address
            ip_string = socket.gethostbyname(host_string)
            ip_address = bytearray(socket.inet_aton(ip_string))

            self._call(IPTagSet(
                connection.chip_x, connection.chip_y, ip_address, ip_tag.port,
                ip_tag.tag, strip=ip_tag.strip_sdp, use_sender=use_sender))

    def __get_connection_list(
            self, connection: Optional[SCAMPConnection] = None,
            board_address: Optional[str] = None) -> List[SCAMPConnection]:
        """
        Get the connections for talking to a board.

        :param SCAMPConnection connection:
            Optional param that directly gives the connection to use.
        :param str board_address:
            Optional param that gives the address of the board to talk to.
        :return: List of length 1 or 0 (the latter only if the search for
            the given board address fails).
        :rtype: list(SCAMPConnection)
        """
        if connection is not None:
            return [connection]
        elif board_address is None:
            return self._scamp_connections

        connection = self._locate_spinnaker_connection_for_board_address(
            board_address)
        if connection is None:
            return []
        return [connection]

    @overrides(Transceiver.set_reverse_ip_tag)
    def set_reverse_ip_tag(self, reverse_ip_tag: ReverseIPTag):
        if reverse_ip_tag.port is None:
            raise SpinnmanInvalidParameterException(
                "reverse_ip_tag.port", "None",
                "The tag port must have been set!")

        if (reverse_ip_tag.port == SCP_SCAMP_PORT or
                reverse_ip_tag.port == UDP_BOOT_CONNECTION_DEFAULT_PORT):
            raise SpinnmanInvalidParameterException(
                "reverse_ip_tag.port", reverse_ip_tag.port,
                "The port number for the reverse IP tag conflicts with"
                f" the SpiNNaker system ports ({SCP_SCAMP_PORT} and "
                f"{UDP_BOOT_CONNECTION_DEFAULT_PORT})")

        # Get the connections - if the tag specifies a connection, use that,
        # otherwise apply the tag to all connections
        connections = self.__get_connection_list(
            None, reverse_ip_tag.board_address)
        if not connections:
            raise SpinnmanInvalidParameterException(
                "reverse_ip_tag", str(reverse_ip_tag),
                "The given board address is not recognised")

        for connection in connections:
            self._call(ReverseIPTagSet(
                connection.chip_x, connection.chip_y,
                reverse_ip_tag.destination_x, reverse_ip_tag.destination_y,
                reverse_ip_tag.destination_p,
                reverse_ip_tag.port, reverse_ip_tag.tag,
                reverse_ip_tag.sdp_port))

    @overrides(Transceiver.clear_ip_tag)
    def clear_ip_tag(self, tag: int, board_address: Optional[str] = None):
        for conn in self.__get_connection_list(board_address=board_address):
            self._call(IPTagClear(conn.chip_x, conn.chip_y, tag))

    @overrides(Transceiver.get_tags)
    def get_tags(self, connection: Optional[SCAMPConnection] = None
                 ) -> Iterable[AbstractTag]:
        all_tags = list()
        for conn in self.__get_connection_list(connection):
            process = GetTagsProcess(self._scamp_connection_selector)
            all_tags.extend(process.get_tags(conn))
        return all_tags

    @overrides(Transceiver.malloc_sdram)
    def malloc_sdram(
            self, x: int, y: int, size: int, app_id: int, tag: int = 0) -> int:
        try:
            process = MallocSDRAMProcess(self._scamp_connection_selector)
            process.malloc_sdram(x, y, size, app_id, tag)
            return process.base_address
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.load_multicast_routes)
    def load_multicast_routes(
            self, x: int, y: int, routes: Collection[MulticastRoutingEntry],
            app_id: int):
        try:
            process = LoadMultiCastRoutesProcess(
                self._scamp_connection_selector)
            process.load_routes(x, y, routes, app_id)
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.load_fixed_route)
    def load_fixed_route(
            self, x: int, y: int, fixed_route: RoutingEntry, app_id: int):
        try:
            process = LoadFixedRouteRoutingEntryProcess(
                self._scamp_connection_selector)
            process.load_fixed_route(x, y, fixed_route, app_id)
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.read_fixed_route)
    def read_fixed_route(self, x: int, y: int, app_id: int) -> RoutingEntry:
        try:
            process = ReadFixedRouteRoutingEntryProcess(
                self._scamp_connection_selector)
            return process.read_fixed_route(x, y, app_id)
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.get_multicast_routes)
    def get_multicast_routes(
            self, x: int, y: int,
            app_id: Optional[int] = None) -> List[MulticastRoutingEntry]:
        try:
            base_address = cast(int, self._get_sv_data(
                x, y, SystemVariableDefinition.router_table_copy_address))
            process = GetMultiCastRoutesProcess(
                self._scamp_connection_selector, app_id)
            return process.get_routes(x, y, base_address)
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.clear_multicast_routes)
    def clear_multicast_routes(self, x: int, y: int):
        try:
            self._call(RouterClear(x, y))
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.get_router_diagnostics)
    def get_router_diagnostics(self, x: int, y: int) -> RouterDiagnostics:
        try:
            process = ReadRouterDiagnosticsProcess(
                self._scamp_connection_selector)
            return process.get_router_diagnostics(x, y)
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.get_scamp_connection_selector)
    def get_scamp_connection_selector(self) -> MostDirectConnectionSelector:
        return self._scamp_connection_selector

    @overrides(Transceiver.set_router_diagnostic_filter)
    def set_router_diagnostic_filter(
            self, x: int, y: int, position: int,
            diagnostic_filter: DiagnosticFilter):
        try:
            self.__set_router_diagnostic_filter(
                x, y, position, diagnostic_filter)
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    def __set_router_diagnostic_filter(
            self, x: int, y: int, position: int,
            diagnostic_filter: DiagnosticFilter):
        data_to_send = diagnostic_filter.filter_word
        if position > NO_ROUTER_DIAGNOSTIC_FILTERS:
            raise SpinnmanInvalidParameterException(
                "position", str(position),
                "the range of the position of a router filter is 0 and 16.")
        if position <= ROUTER_DEFAULT_FILTERS_MAX_POSITION:
            logger.warning(
                "You are planning to change a filter which is set by default. "
                "By doing this, other runs occurring on this machine will be "
                "forced to use this new configuration until the machine is "
                "reset. Please also note that these changes will make the "
                "the reports from ybug not correct. This has been executed "
                "and is trusted that the end user knows what they are doing.")
        memory_position = (
            ROUTER_REGISTER_BASE_ADDRESS + ROUTER_FILTER_CONTROLS_OFFSET +
            position * ROUTER_DIAGNOSTIC_FILTER_SIZE)

        self._call(WriteMemory(
            (x, y, 0), memory_position, _ONE_WORD.pack(data_to_send)))

    @overrides(Transceiver.clear_router_diagnostic_counters)
    def clear_router_diagnostic_counters(self, x: int, y: int):
        try:
            # Clear all
            self._call(WriteMemory(
                (x, y, 0), 0xf100002c, _ONE_WORD.pack(0xFFFFFFFF)))
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @overrides(Transceiver.close)
    def close(self) -> None:
        if self._bmp_connection is not None:
            if get_config_bool("Machine", "turn_off_machine"):
                self._power_off_machine()

        for connection in self._all_connections:
            connection.close()

    @overrides(Transceiver.control_sync)
    def control_sync(self, do_sync: bool):
        self._call(DoSync(do_sync))

    @overrides(Transceiver.update_provenance_and_exit)
    def update_provenance_and_exit(self, x: int, y: int, p: int):
        # Send these signals to make sure the application isn't stuck
        self.send_sdp_message(SDPMessage(
            sdp_header=SDPHeader(
                flags=SDPFlag.REPLY_NOT_EXPECTED,
                destination_port=SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value,
                destination_chip_x=x, destination_chip_y=y, destination_cpu=p),
            data=_ONE_WORD.pack(SDP_RUNNING_MESSAGE_CODES
                                .SDP_UPDATE_PROVENCE_REGION_AND_EXIT.value)))

    @overrides(Transceiver.send_chip_update_provenance_and_exit)
    def send_chip_update_provenance_and_exit(self, x: int, y: int, p: int):
        cmd = SDP_RUNNING_MESSAGE_CODES.SDP_UPDATE_PROVENCE_REGION_AND_EXIT
        port = SDP_PORTS.RUNNING_COMMAND_SDP_PORT

        self.send_sdp_message(SDPMessage(
            SDPHeader(
                flags=SDPFlag.REPLY_NOT_EXPECTED,
                destination_port=port.value, destination_cpu=p,
                destination_chip_x=x, destination_chip_y=y),
            data=_ONE_WORD.pack(cmd.value)))

    def __str__(self) -> str:
        addr = self._scamp_connections[0].remote_ip_address
        n = len(self._all_connections)
        return f"transceiver object connected to {addr} with {n} connections"

    def __repr__(self) -> str:
        return self.__str__()



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import struct
from typing import Callable, List, Sequence
from spinn_utilities.typing.coords import XY, XYP
from spinnman.processes import AbstractMultiConnectionProcess
from spinnman.constants import SYSTEM_VARIABLE_BASE_ADDRESS
from spinnman.model import HeapElement
from spinnman.messages.spinnaker_boot import SystemVariableDefinition
from spinnman.messages.scp.impl.read_memory import ReadMemory, Response
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


HEAP_ADDRESS = SystemVariableDefinition.sdram_heap_address
_ADDRESS = struct.Struct("<I")
_HEAP_POINTER = struct.Struct("<4xI")
_ELEMENT_HEADER = struct.Struct("<II")


class GetHeapProcess(AbstractMultiConnectionProcess[Response]):
    """
    Gets Heap information using the provided connector.

    """
    __slots__ = (
        "_blocks",
        "_heap_address",
        "_next_block_address")

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)

        self._heap_address = 0
        self._next_block_address = 0
        self._blocks: List[HeapElement] = list()

    def _read_heap_address_response(self, response: Response):
        self._heap_address = _ADDRESS.unpack_from(
            response.data, response.offset)[0]

    def _read_heap_pointer(self, response: Response):
        self._next_block_address = _HEAP_POINTER.unpack_from(
            response.data, response.offset)[0]

    def _read_next_block(self, block_address: int, response: Response):
        self._next_block_address, free = _ELEMENT_HEADER.unpack_from(
            response.data, response.offset)
        if self._next_block_address != 0:
            self._blocks.append(HeapElement(
                block_address, self._next_block_address, free))

    def __read_address(
            self, core_coords: XYP, address: int, size: int,
            callback: Callable[[Response], None]):
        with self._collect_responses():
            self._send_request(
                ReadMemory(core_coords, address, size), callback)

    def get_heap(self, chip_coords: XY,
                 pointer: SystemVariableDefinition = HEAP_ADDRESS
                 ) -> Sequence[HeapElement]:
        """
        :param tuple(int,int) chip_coords: x, y
        :param SystemVariableDefinition pointer:
        :rtype: list(HeapElement)
        """
        x, y = chip_coords
        core_coords = (x, y, 0)
        self.__read_address(
            core_coords, SYSTEM_VARIABLE_BASE_ADDRESS + pointer.offset,
            pointer.data_type.value, self._read_heap_address_response)

        self.__read_address(
            core_coords, self._heap_address, 8, self._read_heap_pointer)

        while self._next_block_address != 0:
            self.__read_address(
                core_coords, self._next_block_address, 8,
                functools.partial(
                    self._read_next_block, self._next_block_address))

        return self._blocks



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import AbstractSCPResponse
from spinnman.messages.scp.enums import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.model import VersionInfo


class GetVersionResponse(AbstractSCPResponse):
    """
    An SCP response to a request for the version of software running.
    """
    __slots__ = "_version_info",

    def __init__(self) -> None:
        super().__init__()
        self._version_info: Optional[VersionInfo] = None

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Version", "CMD_VER", result.name)
        self._version_info = VersionInfo(data, offset)

    @property
    def version_info(self) -> Optional[VersionInfo]:
        """
        The version information received.

        :rtype: VersionInfo
        """
        return self._version_info



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The new Spalloc client implementation. This has the notable distinction of
including a proxying system that allows creating a transceiver with access to
the boards of a job despite the client being not within the same firewall/NAT
security domain as the Spalloc-managed service.

The main class in here is :py:class:`~spinnman.spalloc.SpallocClient`.
"""

from .abstract_spalloc_client import AbstractSpallocClient
from .spalloc_job import SpallocJob
from .spalloc_machine import SpallocMachine
from .spalloc_proxied_connection import SpallocProxiedConnection
from .spalloc_eieio_connection import SpallocEIEIOConnection
from .spalloc_eieio_listener import SpallocEIEIOListener
from .spalloc_state import SpallocState
from .spalloc_client import SpallocClient
from .utils import is_server_address

__all__ = (
    "AbstractSpallocClient",
    "is_server_address",
    "SpallocClient",
    "SpallocJob",
    "SpallocMachine",
    "SpallocProxiedConnection",
    "SpallocEIEIOConnection",
    "SpallocEIEIOListener",
    "SpallocState")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class RouterError(Enum):
    """
    Router error flags.
    """
    #: Error packet detected (0x80000000)
    ERROR = 0x80000000
    #: More than one error packet detected (0x40000000)
    OVERFLOW = 0x40000000
    #: Parity Error (0x20000000)
    PARITY = 0x20000000
    #: Framing Error (0x10000000)
    FRAMING = 0x10000000
    #: Timestamp Error (0x08000000)
    TIMESTAMP = 0x08000000



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class DiagnosticFilterDestination(Enum):
    """
    Destination flags for the diagnostic filters.

    .. note::
        Only one has to match for the counter to be incremented.
    """
    #: Destination is to dump the packet
    DUMP = 0
    #: Destination is a local core (but not the monitor core)
    LOCAL = 1
    #: Destination is the local monitor core
    LOCAL_MONITOR = 2
    #: Destination is link 0
    LINK_0 = 3
    #: Destination is link 1
    LINK_1 = 4
    #: Destination is link 2
    LINK_2 = 5
    #: Destination is link 3
    LINK_3 = 6
    #: Destination is link 4
    LINK_4 = 7
    #: Destination is link 5
    LINK_5 = 8



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse

_NNP_FORWARD_RETRY = (1 << 31) | (0x3f << 8) | 0x1A
_NNP_FLOOD_FILL_START = 6


class FloodFillStart(AbstractSCPRequest[CheckOKResponse]):
    """
    A request to start a flood fill of data.
    """
    __slots__ = ()

    def __init__(self, nearest_neighbour_id: int, n_blocks: int,
                 x: Optional[int] = None, y: Optional[int] = None):
        """
        :param int nearest_neighbour_id:
            The ID of the packet, between 0 and 127
        :param int n_blocks:
            The number of blocks of data that will be sent, between 0 and 255
        :param int x: The x-coordinate of the chip to load the data on to. If
            not specified, the data will be loaded on to all chips
        :param int y: The y-coordinate of the chip to load the data on to. If
            not specified, the data will be loaded on to all chips
        """
        key = ((_NNP_FLOOD_FILL_START << 24) | (nearest_neighbour_id << 16) |
               (n_blocks << 8))
        data = 0xFFFF
        if x is not None and y is not None:
            m = ((y & 3) * 4) + (x & 3)
            data = (((x & 0xfc) << 24) + ((y & 0xfc) << 16) +
                    (3 << 16) + (1 << m))

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0,
                destination_chip_x=self.DEFAULT_DEST_X_COORD,
                destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_NNP),
            argument_1=key, argument_2=data, argument_3=_NNP_FORWARD_RETRY)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Flood Fill", "CMD_NNP:NNP_FFS")



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .alloc_free import AllocFree
from .bmp_info import BMPInfo
from .scp_command import SCPCommand
from .iptag_command import IPTagCommand
from .led_action import LEDAction
from .power_command import PowerCommand
from .scp_result import SCPResult
from .signal import Signal, SignalType

__all__ = [
    "AllocFree", "BMPInfo", "SCPCommand",
    "IPTagCommand", "LEDAction",
    "PowerCommand", "SCPResult", "Signal", "SignalType"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Wrappers around socket-related system calls to do exception remapping and
apply some consistency to things.
"""

import logging
import socket
from typing import Optional, Tuple
from spinn_utilities.log import FormatAdapter
from spinnman.exceptions import SpinnmanIOException, SpinnmanTimeoutException

logger = FormatAdapter(logging.getLogger(__name__))


def get_udp_socket() -> socket.socket:
    """
    Wrapper round socket() system call to produce UDP/IPv4 sockets.
    """
    try:
        # Create a UDP Socket
        return socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(
            f"Error setting up socket: {e}") from e


def get_tcp_socket() -> socket.socket:
    """
    Wrapper round socket() system call to produce TCP/IPv4 sockets.

    .. note::
        TCP sockets cannot be used to talk to a SpiNNaker board.
    """
    try:
        # Create a UDP Socket
        return socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(
            f"Error setting up socket: {e}") from e


# pylint: disable=wrong-spelling-in-docstring
def set_receive_buffer_size(sock: socket.socket, size: int):
    """
    Wrapper round setsockopt() system call.
    """
    try:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, size)
    except Exception:  # pylint: disable=broad-except
        # The OS said no, but we might still be able to work right with
        # the defaults. Just warn and hope...
        logger.warning("failed to configure UDP socket to have a large "
                       "receive buffer", exc_info=True)


def bind_socket(sock: socket.socket, host: str, port: int):
    """
    Wrapper round bind() system call.
    """
    try:
        # Bind the socket
        sock.bind((str(host), int(port)))
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(
            f"Error binding socket to {host}:{port}: {e}") from e


def resolve_host(host: str) -> str:
    """
    Wrapper round gethostbyname() system call.
    """
    try:
        return socket.gethostbyname(host)
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(
            f"Error getting IP address for {host}: {e}") from e


def connect_socket(sock: socket.socket, remote_address: str, remote_port: int):
    """
    Wrapper round connect() system call.
    """
    try:
        sock.connect((str(remote_address), int(remote_port)))
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(
            f"Error connecting to {remote_address}:{remote_port}: {e}") from e


def get_socket_address(sock: socket.socket) -> Tuple[str, int]:
    """
    Wrapper round getsockname() system call.
    """
    try:
        addr, port = sock.getsockname()
        # Ensure that a standard address is used for the INADDR_ANY
        # hostname
        if addr is None or addr == "":
            addr = "0.0.0.0"
        return addr, port
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(f"Error querying socket: {e}") from e


def receive_message(
        sock: socket.socket, timeout: Optional[float], size: int) -> bytes:
    """
    Wrapper round recv() system call.
    """
    try:
        sock.settimeout(timeout)
        return sock.recv(size)
    except socket.timeout as e:
        raise SpinnmanTimeoutException("receive", timeout) from e
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(f"Error receiving: {e}") from e


def receive_message_and_address(
        sock: socket.socket, timeout: Optional[float], size: int) -> Tuple[
            bytes, Tuple[str, int]]:
    """
    Wrapper round recvfrom() system call.
    """
    try:
        sock.settimeout(timeout)
        return sock.recvfrom(size)
    except socket.timeout as e:
        raise SpinnmanTimeoutException("receive", timeout) from e
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(f"Error receiving: {e}") from e


def send_message(sock: socket.socket, data: bytes):
    """
    Wrapper round send() system call.
    """
    try:
        return sock.send(data)
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(f"Error sending: {e}") from e


def send_message_to_address(
        sock: socket.socket, data: bytes, address: Tuple[str, int]):
    """
    Wrapper round sendto() system call.
    """
    try:
        return sock.sendto(data, address)
    except Exception as e:  # pylint: disable=broad-except
        raise SpinnmanIOException(f"Error sending: {e}") from e



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class DiagnosticFilterDefaultRoutingStatus(Enum):
    """
    Default routing flags for the diagnostic filters.

    .. note::
        Only one has to match for the counter to be incremented.
    """
    #: Packet is to be default routed
    DEFAULT_ROUTED = 0
    #: Packet is not to be default routed
    NON_DEFAULT_ROUTED = 1



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)
from .application_copy_run_process import ApplicationCopyRunProcess
from .application_run_process import ApplicationRunProcess

from .fixed_connection_selector import FixedConnectionSelector
from .get_heap_process import GetHeapProcess
from .get_cpu_info_process import GetCPUInfoProcess
from .get_exclude_cpu_info_process import GetExcludeCPUInfoProcess
from .get_include_cpu_info_process import GetIncludeCPUInfoProcess
from .get_machine_process import GetMachineProcess
from .get_n_cores_in_state_process import GetNCoresInStateProcess
from .get_routes_process import GetMultiCastRoutesProcess
from .get_tags_process import GetTagsProcess
from .get_version_process import GetVersionProcess
from .load_routes_process import LoadMultiCastRoutesProcess
from .malloc_sdram_process import MallocSDRAMProcess
from .most_direct_connection_selector import MostDirectConnectionSelector
from .read_fixed_route_routing_entry_process import (
    ReadFixedRouteRoutingEntryProcess)
from .load_fixed_route_routing_entry_process import (
    LoadFixedRouteRoutingEntryProcess)
from .read_iobuf_process import ReadIOBufProcess
from .read_memory_process import ReadMemoryProcess
from .read_router_diagnostics_process import ReadRouterDiagnosticsProcess
from .round_robin_connection_selector import RoundRobinConnectionSelector
from .send_single_command_process import SendSingleCommandProcess
from .write_memory_process import WriteMemoryProcess

# Old name
AbstractMultiConnectionProcessConnectionSelector = ConnectionSelector

__all__ = ["AbstractMultiConnectionProcessConnectionSelector",
           "ConnectionSelector",
           "FixedConnectionSelector", "MostDirectConnectionSelector",
           "RoundRobinConnectionSelector",
           "AbstractMultiConnectionProcess",
           "ApplicationRunProcess", "ApplicationCopyRunProcess",
           "GetCPUInfoProcess",
           "GetExcludeCPUInfoProcess", "GetIncludeCPUInfoProcess",
           "GetHeapProcess",
           "GetMachineProcess", "GetMultiCastRoutesProcess",
           "GetNCoresInStateProcess", "GetTagsProcess",
           "GetVersionProcess", "LoadFixedRouteRoutingEntryProcess",
           "LoadMultiCastRoutesProcess", "MallocSDRAMProcess",
           "ReadFixedRouteRoutingEntryProcess", "ReadIOBufProcess",
           "ReadMemoryProcess", "ReadRouterDiagnosticsProcess",
           "SendSingleCommandProcess", "WriteMemoryProcess"]



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable, Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse

_NNP_FORWARD_RETRY = (0x3f << 8) | 0x1A
_NNP_FLOOD_FILL_END = 15
_WAIT_FLAG = 0x1 << 18


class FloodFillEnd(AbstractSCPRequest[CheckOKResponse]):
    """
    A request to start a flood fill of data.
    """
    __slots__ = ()

    def __init__(
            self, nearest_neighbour_id: int, app_id: int = 0,
            processors: Optional[Iterable[int]] = None, wait: bool = False):
        """

        :param int nearest_neighbour_id:
            The ID of the packet, between 0 and 127
        :param int app_id: The application ID to start using the data, between
            16 and 255.  If not specified, no application is started
        :param list(int) processors:
            A list of processors on which to start the application, each
            between 1 and 17. If not specified, no application is started.
        :param bool wait:
            True if the binary should go into a "wait" state before executing
        """
        processor_mask = 0
        if processors is not None:
            for processor in processors:
                processor_mask |= (1 << processor)

        key = (_NNP_FLOOD_FILL_END << 24) | nearest_neighbour_id
        data = (app_id << 24) | processor_mask
        if wait:
            data = data | _WAIT_FLAG

        super().__init__(
            SDPHeader(flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                      destination_cpu=0,
                      destination_chip_x=self.DEFAULT_DEST_X_COORD,
                      destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_NNP),
            argument_1=key, argument_2=data, argument_3=_NNP_FORWARD_RETRY)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Flood Fill", "CMD_NNP:NNP_FFS")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import AllocFree, SCPCommand, SCPResult
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.exceptions import (
    SpinnmanUnexpectedResponseCodeException, SpinnmanInvalidParameterException)

_ONE_WORD = struct.Struct("<I")

FLAG_RETRY_TAG = 4


class _AllocResponse(AbstractSCPResponse):
    """
    An SCP response to a request to allocate space in SDRAM.
    """
    __slots__ = (
        "_base_address",
        "_size")

    def __init__(self, size: int):
        super().__init__()
        self._size = size
        self._base_address: Optional[int] = None

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "SDRAM Allocation", "CMD_ALLOC", result.name)
        self._base_address = _ONE_WORD.unpack_from(data, offset)[0]

        # check that the base address is not null (0 in python case) as
        # this reflects a issue in the command on SpiNNaker side
        if self._base_address == 0:
            raise SpinnmanInvalidParameterException(
                "SDRAM Allocation response base address", self._base_address,
                f"Could not allocate {self._size} bytes of SDRAM")

    @property
    def base_address(self) -> int:
        """
        The base address allocated, or 0 if none.

        :rtype: int
        """
        return self._base_address or 0


class SDRAMAlloc(AbstractSCPRequest[_AllocResponse]):
    """
    An SCP Request to allocate space in the SDRAM space.
    """
    __slots__ = "_size",

    def __init__(self, x: int, y: int, app_id: int, size: int,
                 tag: Optional[int] = None, retry_tag: bool = True):
        """
        :param int x:
            The x-coordinate of the chip to allocate on, between 0 and 255
        :param int y:
            The y-coordinate of the chip to allocate on, between 0 and 255
        :param int app_id: The ID of the application, between 0 and 255
        :param int size: The size in bytes of memory to be allocated
        :param int tag:
            The tag for the SDRAM, a 8-bit (chip-wide) tag that can be
            looked up by a SpiNNaker application to discover the address of
            the allocated block. If `0` then no tag is applied.
        :param bool retry_tag:
            If a tag is used, add a safety check to retry the tag.  This can
            avoid issues with re-allocating memory on a retry message.
        """
        # pylint: disable=too-many-arguments
        extra_flag = 0
        if retry_tag and tag is not None:
            extra_flag = FLAG_RETRY_TAG

        if tag is None:
            tag = 0
        elif not (0 <= tag < 256):
            raise SpinnmanInvalidParameterException(
                "tag",
                "The tag param needs to be between 0 and 255, or None (in "
                "which case 0 will be used by default)", str(tag))

        # pylint: disable=unsupported-binary-operation
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_ALLOC),
            argument_1=(
                (extra_flag << 16) |
                (app_id << 8) |
                AllocFree.ALLOC_SDRAM.value),
            argument_2=size, argument_3=tag)
        self._size = size

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> _AllocResponse:
        return _AllocResponse(self._size)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .get_version_response import GetVersionResponse


class GetVersion(AbstractSCPRequest[GetVersionResponse]):
    """
    An SCP request to read the version of software running on a core.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, p: int):
        """
        :param int x:
            The x-coordinate of the chip to read from, between 0 and 255
        :param int y:
            The y-coordinate of the chip to read from, between 0 and 255
        :param int p: The ID of the processor to read the version from,
            between 0 and 31
        :raise SpinnmanInvalidParameterException:
            * If the chip coordinates are out of range
            * If the processor is out of range
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=p, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_VER))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> GetVersionResponse:
        return GetVersionResponse()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinnman.messages.scp.impl import CountState
from spinnman.messages.scp.enums.scp_result import SCPResult
from .abstract_multi_connection_process import AbstractMultiConnectionProcess

# Timeout for getting core state count; higher due to more waiting needed
GET_CORE_COUNT_TIMEOUT = 2.0


class GetNCoresInStateProcess(AbstractMultiConnectionProcess):
    """
    Gets the state of a core over the provided connection.
    """
    __slots__ = [
        "_n_cores"]

    def __init__(self, connection_selector):
        """
        :param connection_selector:
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector, timeout=GET_CORE_COUNT_TIMEOUT,
                         non_fail_retry_codes={SCPResult.RC_P2P_NOREPLY})
        self._n_cores = 0

    def __handle_response(self, response):
        self._n_cores += response.count

    def get_n_cores_in_state(self, xys, app_id, state):
        """
        :param list(int,int) xys:
        :param int app_id:
        :param int state:
        :rtype: int
        """
        for c_x, c_y in xys:
            self._send_request(
                CountState(c_x, c_y, app_id, state), self.__handle_response)
        self._finish()
        self.check_for_error()

        return self._n_cores



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import cast, Iterable, List, Mapping, Tuple
from spinn_machine import Chip, CoreSubsets, Link, Machine
from spinnman.data import SpiNNManDataView
from spinnman.messages.scp.impl import AppCopyRun
from spinnman.processes import ConnectionSelector
from .abstract_multi_connection_process import AbstractMultiConnectionProcess

APP_COPY_RUN_TIMEOUT = 6.0


def _on_same_board(chip_1: Chip, chip_2: Chip):
    return (chip_1.nearest_ethernet_x == chip_2.nearest_ethernet_x and
            chip_1.nearest_ethernet_y == chip_2.nearest_ethernet_y)


def _get_next_chips(
        chips_done: Mapping[Tuple[int, int], List[Tuple[int, int]]],
        parent_chips: Mapping[Tuple[int, int], List[Chip]],
        machine: Machine) -> Iterable[Chip]:
    """
    Get the chips that are adjacent to the last set of chips, which
    haven't yet been loaded.  Also returned are the links for each chip,
    which gives the link which should be read from to get the data.

    :param dict((int,int),list(int,int)) chips_done:
        The coordinates of chips that have already been done by Ethernet
    :param dict((int, int),~spinn_machine.Chip) parent_chips:
        A dictionary of chip coordinates to chips that use that chip as a
        parent
    :return: A list of next chips to use
    :rtype: list(chip)
    """
    next_chips: List[Chip] = list()
    for eth_chip in chips_done:
        off_board_copy_done = False
        for c_x, c_y in chips_done[eth_chip]:
            chip_xy = machine[c_x, c_y]
            for chip in parent_chips[c_x, c_y]:
                on_same_board = _on_same_board(chip, chip_xy)
                eth = (chip.nearest_ethernet_x, chip.nearest_ethernet_y)
                if (eth not in chips_done or
                        chip not in chips_done[eth]):
                    if on_same_board or not off_board_copy_done:
                        next_chips.append(chip)
                        if not on_same_board:
                            off_board_copy_done = True
                        # Only do one copy from each chip at a time
                        break

    return next_chips


def _compute_parent_chips(
        machine: Machine) -> Mapping[Tuple[int, int], List[Chip]]:
    """
    Compute a dictionary of chip coordinates to list of chips who use that chip
    as a parent in the tree.

    :param ~spinn_machine.Machine machine: The machine to compute the map for
    :rtype: dict((int, int), ~spinn_machine.Chip)
    """
    chip_links: Mapping[Tuple[int, int], List[Chip]] = defaultdict(list)
    for chip in machine.chips:
        if chip.parent_link is not None:
            link = cast(Link, chip.router.get_link(chip.parent_link))
            chip_links[link.destination_x, link.destination_y].append(chip)
    return chip_links


class ApplicationCopyRunProcess(AbstractMultiConnectionProcess):
    """
    Process to start a binary on a subset of cores on a subset of chips
    of a machine, performed by, on each chip, copying the data from
    an adjacent chip and then starting the binary.  This goes to each
    chip in turn, and so detects failures early on, as well as ensuring
    that the copy and execution is done in the case of success i.e. this
    ensures that if all commands are successful, the full binary has been
    copied and started.

    .. note::
        The binary must have been loaded to the boot chip before this is
        called!
    """
    __slots__ = ()

    def __init__(self, next_connection_selector: ConnectionSelector,
                 timeout: float = APP_COPY_RUN_TIMEOUT):
        AbstractMultiConnectionProcess.__init__(
            self, next_connection_selector, timeout=timeout)

    def run(self, size: int, app_id: int, core_subsets: CoreSubsets,
            checksum: int, wait: bool):
        """
        Run the process.

        :param int size: The size of the binary to copy
        :param int app_id: The application id to assign to the running binary
        :param CoreSubsets core_subsets: The cores to load the binary on to
        :param int checksum: The checksum of the data to test against
        :param bool wait:
            Whether to put the binary in "wait" mode or run it straight away
        """
        machine = SpiNNManDataView.get_machine()
        boot_chip = machine.boot_chip
        chips_done: Mapping[Tuple[int, int], List[Tuple[int, int]]] = \
            defaultdict(list)
        chips_done[boot_chip].append((boot_chip))
        parent_chips = _compute_parent_chips(machine)
        next_chips = _get_next_chips(chips_done, parent_chips, machine)

        while next_chips:
            # Do all the chips at the current level
            for chip in next_chips:
                subset = core_subsets.get_core_subset_for_chip(chip.x, chip.y)
                self._send_request(AppCopyRun(
                    chip.x, chip.y, cast(int, chip.parent_link), size, app_id,
                    subset.processor_ids, checksum, wait))
                eth = (chip.nearest_ethernet_x, chip.nearest_ethernet_y)
                chips_done[eth].append(chip)
            self._finish()
            self.check_for_error()
            next_chips = _get_next_chips(chips_done, parent_chips, machine)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional
from spinnman.exceptions import SpinnmanInvalidParameterException
from .spinnaker_boot_op_code import SpinnakerBootOpCode

_PATTERN_HIIII = struct.Struct(">HIIII")
_PATTERN_2xIIII = struct.Struct("2xIIII")
BOOT_MESSAGE_VERSION = 1


class SpinnakerBootMessage(object):
    """
    A message used for booting the board.
    """
    __slots__ = (
        "_data",
        "_offset",
        "_opcode",
        "_operand_1",
        "_operand_2",
        "_operand_3")

    def __init__(self, opcode: SpinnakerBootOpCode,
                 operand_1: int, operand_2: int, operand_3: int,
                 data: Optional[bytes] = None, offset: int = 0):
        """
        :param SpinnakerBootOpCode opcode: The operation of this packet
        :param int operand_1: The first operand
        :param int operand_2: The second operand
        :param int operand_3: The third operand
        :param data: The optional data, up to 256 words
        :type data: bytes or bytearray
        :param int offset: The offset of the valid data
        :raise SpinnmanInvalidParameterException:
            If the opcode is not a valid value
        """
        # pylint: disable=too-many-arguments
        if data is not None and len(data) > (256 * 4):
            raise SpinnmanInvalidParameterException(
                "len(data)", str(len(data)),
                "A boot packet can contain at most 256 words of data")

        self._opcode = opcode
        self._operand_1 = operand_1
        self._operand_2 = operand_2
        self._operand_3 = operand_3
        self._data = data
        self._offset = offset

    @property
    def opcode(self) -> SpinnakerBootOpCode:
        """
        The operation of this packet.

        :rtype: SpinnakerBootOpCode
        """
        return self._opcode

    @property
    def operand_1(self) -> int:
        """
        The first operand.

        :rtype: int
        """
        return self._operand_1

    @property
    def operand_2(self) -> int:
        """
        The second operand.

        :rtype: int
        """
        return self._operand_2

    @property
    def operand_3(self) -> int:
        """
        The third operand.

        :rtype: int
        """
        return self._operand_3

    @property
    def data(self) -> Optional[bytes]:
        """
        The data, or `None` if no data.

        :rtype: bytes or bytearray
        """
        return self._data

    @property
    def bytestring(self) -> bytes:
        """
        The message as a byte-string.

        :rtype: bytes
        """
        data = b""
        if self._data is not None:
            data = self._data[self._offset:]
        return _PATTERN_HIIII.pack(
            BOOT_MESSAGE_VERSION, self._opcode.value,
            self._operand_1, self._operand_2, self._operand_3) + data

    @staticmethod
    def from_bytestring(data: bytes, offset: int) -> 'SpinnakerBootMessage':
        """
        :param bytes data:
        :param int offset:
        :rtype: SpinnakerBootMessage
        """
        (opcode_value, operand_1, operand_2, operand_3) = \
            _PATTERN_2xIIII.unpack_from(data, offset)
        the_data = None
        if len(data) - offset > 0:
            the_data = data
        return SpinnakerBootMessage(
            SpinnakerBootOpCode(opcode_value), operand_1, operand_2, operand_3,
            the_data, offset + 18)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class SignalType(Enum):
    """
    The type of signal, determined by how it is transmitted.
    """
    MULTICAST = 0
    POINT_TO_POINT = 1
    NEAREST_NEIGHBOUR = 2


class Signal(Enum):
    """
    SCP Signals.
    """
    INITIALISE = (0, SignalType.NEAREST_NEIGHBOUR)
    POWER_DOWN = (1, SignalType.NEAREST_NEIGHBOUR)
    STOP = (2, SignalType.NEAREST_NEIGHBOUR)
    START = (3, SignalType.MULTICAST)
    SYNC0 = (4, SignalType.MULTICAST)
    SYNC1 = (5, SignalType.MULTICAST)
    PAUSE = (6, SignalType.MULTICAST)
    CONTINUE = (7, SignalType.MULTICAST)
    EXIT = (8, SignalType.MULTICAST)
    TIMER = (9, SignalType.MULTICAST)
    USER_0 = (10, SignalType.MULTICAST)
    USER_1 = (11, SignalType.MULTICAST)
    USER_2 = (12, SignalType.MULTICAST)
    USER_3 = (13, SignalType.MULTICAST)

    def __new__(cls, *args) -> 'Signal':
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, value: int, signal_type: SignalType) -> None:
        """
        :param int value: The value used for the signal
        :param SignalType signal_type: The "type" of the signal
        """
        assert value >= 0
        self._signal_type = signal_type

    @property
    def signal_type(self) -> SignalType:
        """
        The "type" of the signal

        :rtype: SignalType
        """
        return self._signal_type



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Iterable, Iterator
from spinn_utilities.typing.coords import XYP
from spinnman.model.enums import CPUState
from .cpu_info import CPUInfo


class CPUInfos(object):
    """
    A set of CPU information objects.
    """
    __slots__ = [
        "_cpu_infos"]

    def __init__(self) -> None:
        self._cpu_infos: Dict[XYP, CPUInfo] = dict()

    def add_info(self, cpu_info: CPUInfo):
        """
        Add a info on using its core coordinates.

        :param ~spinnman.model.CPUInfo cpu_info:
        """
        self._cpu_infos[cpu_info.x, cpu_info.y, cpu_info.p] = cpu_info

    def add_infos(self, other, states: Iterable[CPUState]):
        """
        Adds all the infos in the other CPUInfos if the have one of the
        required states

        mainly a support method for Transceiver.add_cpu_information_from_core

        :param CPUInfos other: Another Infos object to merge in
        :param list(CPUState) states:
            Only add if the Info has this state
        """
        # pylint: disable=protected-access
        assert isinstance(other, CPUInfos)
        for info in other._cpu_infos.values():
            if info.state in states:
                self.add_info(info)

    def __iter__(self) -> Iterator[XYP]:
        return iter(self._cpu_infos)

    def __len__(self) -> int:
        """
        The total number of processors that are in these core subsets.
        """
        return len(self._cpu_infos)

    def is_core(self, x: int, y: int, p: int) -> bool:
        """
        Determine if there is a CPU Info for x, y, p.
        """
        return (x, y, p) in self._cpu_infos

    def get_cpu_info(self, x: int, y: int, p: int) -> CPUInfo:
        """
        Get the information for the given core on the given core

        :rtype: CpuInfo
        """
        return self._cpu_infos[x, y, p]

    def infos_for_state(self, state: CPUState) -> 'CPUInfos':
        """
        Creates a new CpuInfos object with Just the Infos that match the state.

        :param ~spinnman.model.enums.CPUState state:
        :return: New Infos object with the filtered infos if any
        :rtype: CPUInfos
        """
        for_state = CPUInfos()
        for info in self._cpu_infos.values():
            if info.state == state:
                for_state.add_info(info)
        return for_state

    def infos_not_in_states(self, states: Iterable[CPUState]) -> 'CPUInfos':
        """
        Creates a new CpuInfos object with Just the Infos that match the state.

        :param iterable(~spinnman.model.enums.CPUState) states:
        :return: New Infos object with the filtered infos if any
        :rtype: CPUInfos
        """
        for_state = CPUInfos()
        for info in self._cpu_infos.values():
            if info.state not in states:
                for_state.add_info(info)
        return for_state

    def get_status_string(self) -> str:
        """
        Get a string indicating the status of the given cores.

        :rtype: str
        """
        break_down = ""
        for core_info in self._cpu_infos.values():
            break_down += core_info.get_status_string()
        return break_down

    def __str__(self) -> str:
        return str([f"{x}, {y}, {p} (ph: {info.physical_cpu_id})"
                    for (x, y, p), info in self._cpu_infos.items()])

    def __repr__(self) -> str:
        return self.__str__()



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path
import time
from typing import List
from spinn_utilities.log import FormatAdapter
from spinn_machine import Machine, Chip
from spinnman.connections.abstract_classes import Connection

logger = FormatAdapter(logging.getLogger(__name__))
_REPORT_NAME = "machine_structure.rpt"


def generate_machine_report(
        report_directory: str, machine: Machine,
        connections: List[Connection]):
    """
    Generate report on the physical structure of the target SpiNNaker machine.

    :param str report_directory: the directory to which reports are stored
    :param ~spinn_machine.Machine machine: the machine python object
    :param list(Connection) connections:
        the list of connections to the machine
    :raise IOError: when a file cannot be opened for some reason
    """
    file_name = os.path.join(report_directory, _REPORT_NAME)
    time_date_string = time.strftime("%c")
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            _write_header(f, time_date_string, machine, connections)
            # TODO: Add further details on the target machine.
            for chip in machine.chips:
                _write_chip_router_report(f, chip)
    except IOError:
        logger.exception(
            "Generate_placement_reports: Can't open file {} for writing.",
            file_name)
        raise


def _write_header(f, timestamp: str, machine: Machine, connections):
    """
    :param str timestamp:
    :param ~spinn_machine.Machine machine:
    :param list(Connection) connections:
    """
    f.write("\t\tTarget SpiNNaker Machine Structure\n")
    f.write("\t\t==================================\n")
    f.write(f"\nGenerated: {timestamp} for target machine '{connections}'\n\n")
    f.write(f"Machine dimensions (in chips) x : {machine.width}  "
            f"y : {machine.height}\n\n")
    f.write("\t\tMachine router information\n")
    f.write("\t\t==========================\n")


def _write_chip_router_report(f, chip: Chip):
    """
    :param ~spinn_machine.Chip chip:
    """
    f.write(f"\nInformation for chip {chip.x}:{chip.y}\n")
    f.write("Neighbouring chips\n"
            f"{chip.router.get_neighbouring_chips_coords()}\n")
    f.write("Router list of links for this chip are: \n")
    for link in chip.router.links:
        f.write(f"\t{link}\n")
    f.write("\t\t==========================\n")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from functools import partial
from typing import List, Optional

from spinn_machine import MulticastRoutingEntry, RoutingEntry
from spinnman.messages.scp.impl.read_memory import ReadMemory, Response
from spinnman.constants import UDP_MESSAGE_MAX_SIZE

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)

# There are 1024 entries in a routing table
_N_ENTRIES = 1024

# 16 entries fit in a 256-byte read
_ENTRIES_PER_READ = 16

# 64 reads of 16 entries are required for 1024 entries
_N_READS = 64

_ROUTE_ENTRY_PATTERN = struct.Struct("<2xBxIII")


class GetMultiCastRoutesProcess(AbstractMultiConnectionProcess[Response]):
    """
    A process for reading the multicast routing table of a SpiNNaker chip.
    """
    __slots__ = (
        "_app_id",
        "_entries")

    def __init__(self, connection_selector: ConnectionSelector,
                 app_id: Optional[int] = None):
        """
        :param ConnectionSelector connection_selector:
        :param int app_id:
        """
        super().__init__(connection_selector)
        self._entries: List[Optional[MulticastRoutingEntry]] = \
            [None] * _N_ENTRIES
        self._app_id = app_id

    def _add_routing_entry(
            self, route_no: int, offset: int, app_id: int, route: int,
            key: int, mask: int):
        # pylint: disable=too-many-arguments
        if route >= 0xFF000000:
            return
        if self._app_id is not None and self._app_id != app_id:
            return

        self._entries[route_no + offset] = MulticastRoutingEntry(
            key, mask, RoutingEntry(spinnaker_route=route))

    def __handle_response(self, offset: int, response: Response):
        for route_no in range(_ENTRIES_PER_READ):
            entry = _ROUTE_ENTRY_PATTERN.unpack_from(
                response.data,
                response.offset + route_no * _ROUTE_ENTRY_PATTERN.size)
            self._add_routing_entry(route_no, offset, *entry)

    def get_routes(self, x: int, y: int,
                   base_address: int) -> List[MulticastRoutingEntry]:
        """
        :param int x:
        :param int y:
        :param int base_address:
        :rtype: list(~spinn_machine.MulticastRoutingEntry)
        """
        # Create the read requests
        offset = 0
        scamp_coords = x, y, 0
        with self._collect_responses():
            for _ in range(_N_READS):
                self._send_request(
                    ReadMemory(
                        scamp_coords, base_address + (offset * 16),
                        UDP_MESSAGE_MAX_SIZE),
                    partial(self.__handle_response, offset))
                offset += _ENTRIES_PER_READ

        return [entry for entry in self._entries if entry is not None]



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
API of the client for the Spalloc web service.
"""

import struct
from typing import Dict, Iterable, Tuple, Optional
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from .spalloc_machine import SpallocMachine
from .spalloc_job import SpallocJob

_ONE_SHORT = struct.Struct("<H")
_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP: bytes = b'\0\0'
_NUM_UPDATE_TAG_TRIES = 3
_UPDATE_TAG_TIMEOUT = 1.0


class AbstractSpallocClient(object, metaclass=AbstractBase):
    """
    The API exported by the Spalloc Client.
    """
    __slots__ = ()

    @abstractmethod
    def list_machines(self) -> Dict[str, SpallocMachine]:
        """
        Get the machines supported by the server.

        :return:
            Mapping from machine names to handles for working with a machine.
        :rtype: dict(str,SpallocMachine)
        """
        raise NotImplementedError

    @abstractmethod
    def list_jobs(self, deleted: bool = False) -> Iterable[SpallocJob]:
        """
        Get the jobs known to the server.

        :param bool deleted: Whether to include deleted jobs.
        :return: The jobs known to the server.
        :rtype: ~typing.Iterable(SpallocJob)
        """
        raise NotImplementedError

    @abstractmethod
    def create_job(
            self, num_boards: int = 1, machine_name: Optional[str] = None,
            keepalive: int = 45) -> SpallocJob:
        """
        Create a job with a specified number of boards.

        :param int num_boards:
            How many boards to ask for (defaults to 1)
        :param str machine_name:
            Which machine to run on? If omitted, the service's machine tagged
            with ``default`` will be used.
        :param int keepalive:
            After how many seconds of no activity should a job become eligible
            for automatic pruning?
        :return: A handle for monitoring and interacting with the job.
        :rtype: SpallocJob
        """
        raise NotImplementedError

    @abstractmethod
    def create_job_rect(
            self, width: int, height: int, machine_name: Optional[str] = None,
            keepalive: int = 45) -> SpallocJob:
        """
        Create a job with a rectangle of boards.

        :param int width:
            The width of rectangle to request
        :param int height:
            The height of rectangle to request
        :param str machine_name:
            Which machine to run on? If omitted, the service's machine tagged
            with ``default`` will be used.
        :param int keepalive:
            After how many seconds of no activity should a job become eligible
            for automatic pruning?
        :return: A handle for monitoring and interacting with the job.
        :rtype: SpallocJob
        """
        raise NotImplementedError

    @abstractmethod
    def create_job_board(
            self, triad: Optional[Tuple[int, int, int]] = None,
            physical: Optional[Tuple[int, int, int]] = None,
            ip_address: Optional[str] = None,
            machine_name: Optional[str] = None,
            keepalive: int = 45) -> SpallocJob:
        """
        Create a job with a specific board. At least one of ``triad``,
        ``physical`` and ``ip_address`` must be not ``None``.

        :param tuple(int,int,int) triad:
            The logical coordinate of the board to request
        :param tuple(int,int,int) physical:
            The physical coordinate of the board to request
        :param str ip_address:
            The IP address of the board to request
        :param str machine_name:
            Which machine to run on? If omitted, the service's machine tagged
            with ``default`` will be used.
        :param int keepalive:
            After how many seconds of no activity should a job become eligible
            for automatic pruning?
        :return: A handle for monitoring and interacting with the job.
        :rtype: SpallocJob
        """
        raise NotImplementedError

    @abstractmethod
    def create_job_rect_at_board(
            self, width: int, height: int,
            triad: Optional[Tuple[int, int, int]] = None,
            physical: Optional[Tuple[int, int, int]] = None,
            ip_address: Optional[str] = None,
            machine_name: Optional[str] = None, keepalive: int = 45,
            max_dead_boards: int = 0) -> SpallocJob:
        """
        Create a job with a rectangle of boards starting at a specific board.
        At least one of ``triad``, ``physical`` and ``ip_address`` must be not
        ``None``.

        :param int width:
            The width of rectangle to request
        :param int height:
            The height of rectangle to request
        :param tuple(int,int,int) triad:
            The logical coordinate of the board to request
        :param tuple(int,int,int) physical:
            The physical coordinate of the board to request
        :param str ip_address:
            The IP address of the board to request
        :param str machine_name:
            Which machine to run on? If omitted, the service's machine tagged
            with ``default`` will be used.
        :param int keepalive:
            After how many seconds of no activity should a job become eligible
            for automatic pruning?
        :param int max_dead_boards:
            How many dead boards can be included.
        :return: A handle for monitoring and interacting with the job.
        :rtype: SpallocJob
        """
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import logging
import sys
from types import TracebackType
from typing import (
    Callable, Dict, Generator, Generic, List, Optional, TypeVar, cast, Set)

from typing_extensions import Self, TypeAlias

from spinn_utilities.log import FormatAdapter

from spinnman.connections import SCPRequestPipeLine
from spinnman.connections.udp_packet_connections import SCAMPConnection
from spinnman.constants import SCP_TIMEOUT, N_RETRIES
from spinnman.exceptions import (
    SpinnmanGenericProcessException, SpinnmanGroupedProcessException)
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums.scp_result import SCPResult

from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)

#: Type of responses.
#: :meta private:
R = TypeVar("R", bound=AbstractSCPResponse)
#: Error handling callback.
#: :meta private:
ECB: TypeAlias = Callable[
    [AbstractSCPRequest[R], Exception, TracebackType, SCAMPConnection], None]

logger = FormatAdapter(logging.getLogger(__name__))


class AbstractMultiConnectionProcess(Generic[R]):
    """
    A process for talking to SpiNNaker efficiently that uses multiple
    connections in communication if relevant.
    """
    __slots__ = (
        "_error_requests",
        "_exceptions",
        "_tracebacks",
        "_connections",
        "_intermediate_channel_waits",
        "_n_channels",
        "_n_retries",
        "_non_fail_retry_codes",
        "_conn_selector",
        "_scp_request_pipelines",
        "_timeout")

    def __init__(self, next_connection_selector: ConnectionSelector,
                 n_retries: int = N_RETRIES, timeout: float = SCP_TIMEOUT,
                 n_channels: int = 8, intermediate_channel_waits: int = 7,
                 non_fail_retry_codes: Optional[Set[SCPResult]] = None):
        """
        :param ConnectionSelector next_connection_selector:
            How to choose the connection.
        :param int n_retries:
            The number of retries of a message to use. Passed to
            :py:class:`SCPRequestPipeLine`
        :param float timeout:
            The timeout, in seconds. Passed to :py:class:`SCPRequestPipeLine`
        :param int n_channels:
            The maximum number of channels to use when talking to a particular
            SCAMP instance. Passed to :py:class:`SCPRequestPipeLine`
        :param int intermediate_channel_waits:
            The maximum number of outstanding message/reply pairs to have on a
            particular connection. Passed to :py:class:`SCPRequestPipeLine`
        :param Optional[Set[SCPResult]] non_fail_retry_codes:
            Optional set of responses that result in retry but after retrying
            don't then result in failure even if returned on the last call.
        """
        self._exceptions: List[Exception] = []
        self._tracebacks: List[TracebackType] = []
        self._error_requests: List[AbstractSCPRequest[R]] = []
        self._connections: List[SCAMPConnection] = []
        self._scp_request_pipelines: Dict[
            SCAMPConnection, SCPRequestPipeLine[R]] = dict()
        self._n_retries = n_retries
        self._timeout = timeout
        self._n_channels = n_channels
        self._intermediate_channel_waits = intermediate_channel_waits
        self._conn_selector = next_connection_selector
        self._non_fail_retry_codes = non_fail_retry_codes

    def _send_request(self, request: AbstractSCPRequest[R],
                      callback: Optional[Callable[[R], None]] = None,
                      error_callback: Optional[ECB] = None):
        if error_callback is None:
            error_callback = self._receive_error
        connection = self._conn_selector.get_next_connection(request)
        if connection not in self._scp_request_pipelines:
            self._scp_request_pipelines[connection] = SCPRequestPipeLine(
                connection, n_retries=self._n_retries,
                packet_timeout=self._timeout,
                n_channels=self._n_channels,
                intermediate_channel_waits=self._intermediate_channel_waits,
                non_fail_retry_codes=self._non_fail_retry_codes)
        self._scp_request_pipelines[connection].send_request(
            request, callback, error_callback)

    def _receive_error(
            self, request: AbstractSCPRequest[R], exception: Exception,
            tb: TracebackType, connection: SCAMPConnection):
        self._error_requests.append(request)
        self._exceptions.append(exception)
        self._tracebacks.append(tb)
        self._connections.append(connection)

    def is_error(self) -> bool:
        """
        Checks if any errors have been cached.

        :rtype: bool
        """
        return bool(self._exceptions)

    def _finish(self) -> None:
        for request_pipeline in self._scp_request_pipelines.values():
            request_pipeline.finish()

    @contextlib.contextmanager
    def _collect_responses(
            self, *, print_exception: bool = False,
            check_error: bool = True) -> Generator[Self, None, None]:
        """
        A simple context for calls. Lets you do this::

            with self._collect_responses():
                self._send_request(...)

        instead of this::

            self._send_request(...)
            self._finish()
            self.check_for_error()

        :param bool print_exception:
            Whether to log errors as well as raising
        :param bool check_error:
            Whether to check for errors; if not, caller must handle
        """
        yield self
        for request_pipeline in self._scp_request_pipelines.values():
            request_pipeline.finish()
        if check_error and self._exceptions:
            self.check_for_error(print_exception=print_exception)

    @property
    def connection_selector(self) -> ConnectionSelector:
        """
        The connection selector of the process.

        :rtype: ConnectionSelector
        """
        return self._conn_selector

    def check_for_error(self, print_exception: bool = False):
        """
        Check if any errors have been cached and raises them

        if print_exception it also logs the error.

        :param bool print_exception:
        :raises SpinnmanGenericProcessException: If any was found
        """
        if len(self._exceptions) == 1:
            exc_info = sys.exc_info()
            sdp_header = self._error_requests[0].sdp_header
            phys_p = sdp_header.get_physical_cpu_id()

            if print_exception:
                connection = self._connections[0]
                logger.error(
                    "failure in request to board {} with ethernet chip "
                    "({}, {}) for chip ({}, {}, {}({}))",
                    connection.remote_ip_address, connection.chip_x,
                    connection.chip_y, sdp_header.destination_chip_x,
                    sdp_header.destination_chip_y, sdp_header.destination_cpu,
                    phys_p)

            raise SpinnmanGenericProcessException(
                self._exceptions[0], cast(TracebackType, exc_info[2]),
                sdp_header.destination_chip_x,
                sdp_header.destination_chip_y, sdp_header.destination_cpu,
                phys_p, self._tracebacks[0])
        elif self._exceptions:
            ex = SpinnmanGroupedProcessException(
                self._error_requests, self._exceptions, self._tracebacks,
                self._connections)
            if print_exception:
                logger.error("{}", str(ex))
            raise ex



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.messages.sdp import SDPHeader
from spinnman.messages.scp import SCPResponseHeader

# The offset of the header from the start of a received packet
# (8 bytes of SDP header)
_SCP_HEADER_OFFSET = 8

# The offset of the data from the start of a received packet
# (8 bytes of SDP header + 4 bytes of SCP header)
_SCP_DATA_OFFSET = 12


class AbstractSCPResponse(object, metaclass=AbstractBase):
    """
    Represents an abstract SCP Response.
    """
    __slots__ = (
        "_scp_response_header",
        "_sdp_header")

    def __init__(self) -> None:
        self._sdp_header: Optional[SDPHeader] = None
        self._scp_response_header: Optional[SCPResponseHeader] = None

    def read_bytestring(self, data: bytes, offset: int):
        """
        Reads a packet from a byte-string of data.

        :param bytes data: The byte-string to be read
        :param int offset:
            The offset in the data from which the response should be read
        """
        self._sdp_header = SDPHeader.from_bytestring(data, offset)
        self._scp_response_header = SCPResponseHeader.from_bytestring(
            data, _SCP_HEADER_OFFSET + offset)
        self.read_data_bytestring(data, _SCP_DATA_OFFSET + offset)

    @abstractmethod
    def read_data_bytestring(self, data: bytes, offset: int):
        """
        Reads the remainder of the data following the header.

        :param bytes data: The byte-string to read from
        :param int offset: The offset into the data after the headers
        """
        raise NotImplementedError

    @property
    def sdp_header(self) -> SDPHeader:
        """
        The SDP header from the response.

        :rtype: SDPHeader
        """
        assert self._sdp_header is not None, "no response read"
        return self._sdp_header

    @property
    def scp_response_header(self) -> SCPResponseHeader:
        """
        The SCP header from the response.

        :rtype: SCPResponseHeader
        """
        assert self._scp_response_header is not None, "no response read"
        return self._scp_response_header




# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional
from spinnman.messages.scp.impl import SDRAMDeAlloc
from spinnman.processes import AbstractMultiConnectionProcess
from spinnman.processes import ConnectionSelector


class DeAllocSDRAMProcess(AbstractMultiConnectionProcess):
    """
    .. warning::
        This class is currently deprecated and untested as there is no
        known use except for Transceiver.free_sdram and free_sdram_by_app_id
        which are both themselves deprecated.
    """
    __slots__ = ("_no_blocks_freed", )

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)
        self._no_blocks_freed: Optional[int] = None

    def de_alloc_all_app_sdram(self, x: int, y: int, app_id: int):
        """
        :param int x:
        :param int y:
        :param int app_id:
        :param base_address:
        :type base_address: int or None
        """
        # deallocate space in the SDRAM
        with self._collect_responses():
            self._send_request(SDRAMDeAlloc(x, y, app_id=app_id),
                               callback=self.__handle_sdram_alloc_response)

    def de_alloc_sdram(self, x: int, y: int, base_address: int):
        """
        :param int x:
        :param int y:
        :param int base_address:
        """
        with self._collect_responses():
            self._send_request(SDRAMDeAlloc(x, y, base_address=base_address),
                               callback=None)

    def __handle_sdram_alloc_response(self, response):
        self._no_blocks_freed = response.number_of_blocks_freed

    @property
    def no_blocks_freed(self) -> Optional[int]:
        """
        :rtype: int
        """
        return self._no_blocks_freed



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.eieio import AbstractEIEIOMessage
from spinnman.messages.eieio.command_messages import EIEIOCommandHeader


class EIEIOCommandMessage(AbstractEIEIOMessage):
    """
    An EIEIO command message.
    """
    __slots__ = (
        "_data",
        "_eieio_command_header",
        "_offset")

    def __init__(self, eieio_command_header: EIEIOCommandHeader,
                 data: Optional[bytes] = None, offset: int = 0):
        """
        :param EIEIOCommandHeader eieio_command_header:
            The header of the message
        :param bytes data: Optional incoming data
        :param int offset: Offset into the data where valid data begins
        """
        # The header
        self._eieio_command_header = eieio_command_header

        # The data
        self._data = data
        self._offset = offset

    @property
    @overrides(AbstractEIEIOMessage.eieio_header)
    def eieio_header(self) -> EIEIOCommandHeader:
        """
        Gets the eieio_header passed into the init.

        :rtype: EIEIOCommandHeader
        """
        return self._eieio_command_header

    @property
    def data(self) -> Optional[bytes]:
        """
        Gets the data passed into the init (if applicable).

        :rtype: bytes or None
        """
        return self._data

    @property
    def offset(self) -> int:
        """
        Gets the offset passed into the init

        :rtype: int
        """
        return self._offset

    @staticmethod
    def from_bytestring(command_header: EIEIOCommandHeader, data: bytes,
                        offset: int) -> "EIEIOCommandMessage":
        """
        Creates an EIEIOCommandMessage based on the supplied information.

        :param EIEIOCommandHeader command_header:
        :param bytes data:
        :param int offset:
        :rtype: EIEIOCommandMessage
        """
        return EIEIOCommandMessage(command_header, data, offset)

    @property
    def bytestring(self) -> bytes:
        """
        The eieio_command_header passed into the init as a byte string.

        :rtype: bytes
        """
        return self._eieio_command_header.bytestring

    @staticmethod
    def get_min_packet_length() -> int:
        """
        Gets the min packet length for this type.

        :rtype: int
        """
        return 2

    def __str__(self):
        return f"EIEIOCommandMessage:{self._eieio_command_header}"

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
API of the client for the Spalloc web service.
"""

import struct
import time
from typing import Callable, Optional
from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.overrides import overrides
from spinnman.connections.abstract_classes import Listenable
from spinnman.connections.udp_packet_connections import BootConnection
from spinnman.connections.udp_packet_connections.boot_connection import (
    _ANTI_FLOOD_DELAY)
from spinnman.messages.spinnaker_boot import SpinnakerBootMessage
from .spalloc_proxied_connection import SpallocProxiedConnection

_ONE_SHORT = struct.Struct("<H")
_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP: bytes = b'\0\0'
_NUM_UPDATE_TAG_TRIES = 3
_UPDATE_TAG_TIMEOUT = 1.0


class SpallocBootConnection(
        BootConnection, SpallocProxiedConnection,
        Listenable[SpinnakerBootMessage], metaclass=AbstractBase):
    """
    The socket interface supported by proxied boot sockets. The socket will
    always be talking to the root board of a job.
    This emulates a BootConnection.
    """
    __slots__ = ()

    @overrides(BootConnection.send_boot_message)
    def send_boot_message(self, boot_message: SpinnakerBootMessage):
        self.send(boot_message.bytestring)

        # Sleep between messages to avoid flooding the machine
        time.sleep(_ANTI_FLOOD_DELAY)

    @overrides(BootConnection.receive_boot_message)
    def receive_boot_message(
            self, timeout: Optional[float] = None) -> SpinnakerBootMessage:
        data = self.receive(timeout)
        return SpinnakerBootMessage.from_bytestring(data, 0)

    @overrides(Listenable.get_receive_method)
    def get_receive_method(  # type: ignore[override]
            self) -> Callable[[], SpinnakerBootMessage]:
        return self.receive_boot_message



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Container
from spinn_utilities.overrides import overrides
from spinnman.model import CPUInfo
from spinnman.model.enums import CPUState
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)
from .get_cpu_info_process import GetCPUInfoProcess


class GetIncludeCPUInfoProcess(GetCPUInfoProcess):
    """
    Gets the CPU for processors over the provided connection.

    This class returns only infos for the requested states.
    """
    __slots__ = ("__states", )

    def __init__(self, connection_selector: ConnectionSelector,
                 states: Container[CPUState]):
        """
        :param connection_selector:
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        :param iterable(CPUState) states:
            The states for which info is required.
        """
        super().__init__(connection_selector)
        self.__states = states

    @overrides(GetCPUInfoProcess._is_desired)
    def _is_desired(self, cpu_info: CPUInfo) -> bool:
        return cpu_info.state in self.__states



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cpu_state import CPUState
from .diagnostic_filter_default_routing_status import (
    DiagnosticFilterDefaultRoutingStatus)
from .diagnostic_filter_destination import DiagnosticFilterDestination
from .diagnostic_filter_emergency_routing_status import (
    DiagnosticFilterEmergencyRoutingStatus)
from .diagnostic_filter_packet_type import DiagnosticFilterPacketType
from .diagnostic_filter_payload_status import DiagnosticFilterPayloadStatus
from .diagnostic_filter_source import DiagnosticFilterSource
from .executable_type import ExecutableType
from .mailbox_command import MailboxCommand
from .p2p_table_route import P2PTableRoute
from .run_time_error import RunTimeError
from .router_error import RouterError
from .sdp_ports import SDP_PORTS
from .sdp_running_message_codes import SDP_RUNNING_MESSAGE_CODES
from .user_registers import UserRegister

__all__ = ["CPUState", "DiagnosticFilterDefaultRoutingStatus",
           "DiagnosticFilterDestination",
           "DiagnosticFilterEmergencyRoutingStatus",
           "DiagnosticFilterPacketType", "DiagnosticFilterPayloadStatus",
           "DiagnosticFilterSource", "ExecutableType", "MailboxCommand",
           "P2PTableRoute", "RouterError", "RunTimeError", "SDP_PORTS",
           "SDP_RUNNING_MESSAGE_CODES", "UserRegister"]



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
API of the client for the Spalloc web service.
"""
from typing import Optional
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.connections.abstract_classes import Listenable


class SpallocProxiedConnection(Listenable, metaclass=AbstractBase):
    """
    Base class for connections proxied via Spalloc.
    """
    __slots__ = ()

    @abstractmethod
    def send(self, data: bytes):
        """
        Send a message on an open socket.

        :param data: The message to send.
        """
        raise NotImplementedError

    @abstractmethod
    def receive(self, timeout: Optional[float] = None) -> bytes:
        """
        Receive a message on an open socket. Will block until a message is
        received.

        :param float timeout:
            How long to wait for a message to be received before timing out.
            If `None`, will wait indefinitely (or until the connection is
            closed).
        :type timeout: int or float or None
        :return: The received message.
        :raises SpinnmanTimeoutException:
            If a timeout happens
        """
        raise NotImplementedError



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from logging import getLogger
from json.decoder import JSONDecodeError
import re
from typing import Dict, Tuple, cast, Optional
import websocket  # type: ignore

import requests

from spinn_utilities.log import FormatAdapter
from spinn_utilities.typing.json import JsonObject
from spinnman.exceptions import SpallocException

from .utils import clean_url

logger = FormatAdapter(getLogger(__name__))
#: The name of the session cookie issued by Spring Security
_SESSION_COOKIE = "JSESSIONID"
#: Enable detailed debugging by setting to True
_debug_pretty_print = False


def _may_renew(method):
    def pp_req(request: requests.PreparedRequest):
        """
        :param ~requests.PreparedRequest request:
        """
        print(">>>>>>>>>>>START>>>>>>>>>>>\n")
        print(f"{request.method} {request.url}")
        # pylint: disable=consider-using-f-string
        print('\r\n'.join('{}: {}'.format(*kv)
                          for kv in request.headers.items()))
        if request.body:
            print(request.body)

    def pp_resp(response: requests.Response):
        """
        :param ~requests.Response response:
        """
        # pylint: disable=consider-using-f-string
        print('{}\n{}\r\n{}\r\n\r\n{}'.format(
            '<<<<<<<<<<<START<<<<<<<<<<<',
            str(response.status_code) + " " + response.reason,
            '\r\n'.join('{}: {}'.format(*kv)
                        for kv in response.headers.items()),
            # Assume we only get textual responses
            str(response.content, "UTF-8") if response.content else ""))

    @wraps(method)
    def call(self, *args, **kwargs):
        renew_count = 0
        while True:
            r = method(self, *args, **kwargs)
            if _debug_pretty_print:
                pp_req(r.request)
                pp_resp(r)
            if _SESSION_COOKIE in r.cookies:
                # pylint: disable=protected-access
                self._session_id = r.cookies[_SESSION_COOKIE]
            if r.status_code != 401 or not renew_count:
                return r
            self.renew()
            renew_count += 1

    return call


class Session:
    """
    Manages session credentials for the Spalloc client.

    .. warning::
        This class does not present a stable API for public consumption.
    """
    __slots__ = (
        "__login_form_url", "__login_submit_url", "__srv_base", "_service_url",
        "__username", "__password", "__token",
        "_session_id", "__csrf", "__csrf_header")

    def __init__(
            self, service_url: str,
            username: Optional[str] = None, password: Optional[str] = None,
            token: Optional[str] = None,
            session_credentials: Optional[
                Tuple[Dict[str, str], Dict[str, str]]] = None):
        """
        :param str service_url: The reference to the service.
            *Should not* include a username or password in it.
        :param str username: The user name to use
        :param str password: The password to use
        :param str token: The bearer token to use
        """
        url = clean_url(service_url)
        self.__login_form_url = url + "system/login.html"
        self.__login_submit_url = url + "system/perform_login"
        self._service_url = url
        self.__srv_base = url + "srv/spalloc/"
        self.__username = username
        self.__password = password
        self.__token = token
        if session_credentials:
            cookies, headers = session_credentials
            if _SESSION_COOKIE in cookies:
                self._session_id = cookies[_SESSION_COOKIE]
            for key, value in headers.items():
                if key == "Authorization":
                    # TODO: extract this?
                    pass
                else:
                    self.__csrf_header = key
                    self.__csrf = value

    def __handle_error_or_return(self, response: requests.Response):
        code = response.status_code
        if code >= 200 and code < 400:
            return response
        result = response.content
        raise ValueError(f"Unexpected response from server {code}\n"
                         f"    {str(result)}")

    @_may_renew
    def get(self, url: str, timeout: int = 10, **kwargs) -> requests.Response:
        """
        Do an HTTP ``GET`` in the session.

        :param str url:
        :param int timeout:
        :rtype: ~requests.Response
        :raise ValueError: If the server rejects a request
        """
        params = kwargs if kwargs else None
        cookies = {_SESSION_COOKIE: self._session_id}
        r = requests.get(url, params=params, cookies=cookies,
                         allow_redirects=False, timeout=timeout)
        logger.debug("GET {} returned {}", url, r.status_code)
        return self.__handle_error_or_return(r)

    @_may_renew
    def post(self, url: str, json_dict: dict, timeout: int = 10,
             **kwargs) -> requests.Response:
        """
        Do an HTTP ``POST`` in the session.

        :param str url:
        :param int timeout:
        :param dict json_dict:
        :rtype: ~requests.Response
        :raise ValueError: If the server rejects a request
        """
        params = kwargs if kwargs else None
        cookies, headers = self._credentials
        r = requests.post(url, params=params, json=json_dict,
                          cookies=cookies, headers=headers,
                          allow_redirects=False, timeout=timeout)
        logger.debug("POST {} returned {}", url, r.status_code)
        return self.__handle_error_or_return(r)

    @_may_renew
    def put(self, url: str, data: str, timeout: int = 10,
            **kwargs) -> requests.Response:
        """
        Do an HTTP ``PUT`` in the session. Puts plain text *OR* JSON!

        :param str url:
        :param str data:
        :param int timeout:
        :rtype: ~requests.Response
        :raise ValueError: If the server rejects a request
        """
        params = kwargs if kwargs else None
        cookies, headers = self._credentials
        if isinstance(data, str):
            headers["Content-Type"] = "text/plain; charset=UTF-8"
        r = requests.put(url, params=params, data=data,
                         cookies=cookies, headers=headers,
                         allow_redirects=False, timeout=timeout)
        logger.debug("PUT {} returned {}", url, r.status_code)
        return self.__handle_error_or_return(r)

    @_may_renew
    def delete(self, url: str, timeout: int = 10,
               **kwargs) -> requests.Response:
        """
        Do an HTTP ``DELETE`` in the session.

        :param str url:
        :rtype: ~requests.Response
        :raise ValueError: If the server rejects a request
        """
        params = kwargs if kwargs else None
        cookies, headers = self._credentials
        r = requests.delete(url, params=params, cookies=cookies,
                            headers=headers, allow_redirects=False,
                            timeout=timeout)
        logger.debug("DELETE {} returned {}", url, r.status_code)
        return self.__handle_error_or_return(r)

    def renew(self) -> JsonObject:
        """
        Renews the session, logging the user into it so that state modification
        operations can be performed.

        :returns: Description of the root of the service, without CSRF data
        :rtype: dict
        :raises SpallocException:
            If the session cannot be renewed.
        """
        if self.__token:
            r = requests.get(
                self.__login_form_url,
                headers={"Authorization": f"Bearer {self.__token}"},
                allow_redirects=False, timeout=10)
            if not r.ok:
                raise SpallocException(
                    f"Could not renew session: {cast(str, r.content)}")
            self._session_id = r.cookies[_SESSION_COOKIE]
        else:
            # Step one: a temporary session so we can log in
            csrf_matcher = re.compile(
                r"""<input type="hidden" name="_csrf" value="(.*)" />""")
            r = requests.get(self.__login_form_url, allow_redirects=False,
                             timeout=10)
            logger.debug("GET {} returned {}",
                         self.__login_form_url, r.status_code)
            m = csrf_matcher.search(r.text)
            if not m:
                msg = ("Could not establish temporary session to "
                       f"{self._service_url} for user {self.__username} ")
                if self.__password is None:
                    msg += "with a no password"
                else:
                    msg += f"with a {len(self.__password)} character password."
                raise SpallocException(msg)
            csrf = m.group(1)
            session = r.cookies[_SESSION_COOKIE]

            # Step two: actually do the log in
            form = {
                "_csrf": csrf,
                "username": self.__username,
                "password": self.__password,
                "submit": "submit"
            }
            # NB: returns redirect that sets a cookie
            r = requests.post(self.__login_submit_url,
                              cookies={_SESSION_COOKIE: session},
                              allow_redirects=False,
                              data=form, timeout=10)
            logger.debug("POST {} returned {}",
                         self.__login_submit_url, r.status_code)
            try:
                self._session_id = r.cookies[_SESSION_COOKIE]
            except KeyError as e:
                try:
                    json_error = r.json()
                    if 'message' in json_error:
                        error = json_error['message']
                    else:
                        error = str(json_error)
                except JSONDecodeError:
                    error = r.raw
                raise SpallocException(f"Unable to login: {error}") from e

        # Step three: get the basic service data and new CSRF token
        obj: JsonObject = self.get(self.__srv_base).json()
        self.__csrf_header = cast(str, obj["csrf-header"])
        self.__csrf = cast(str, obj["csrf-token"])
        del obj["csrf-header"]
        del obj["csrf-token"]
        return obj

    @property
    def _credentials(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        The credentials for requests. *Serializable.*
        """
        cookies = {_SESSION_COOKIE: self._session_id}
        headers = {self.__csrf_header: self.__csrf}
        if self.__token:
            # This would be better off done once per session only
            headers["Authorization"] = f"Bearer {self.__token}"
        return cookies, headers

    def websocket(
            self, url: str, header: Optional[dict] = None,
            cookie: Optional[str] = None, **kwargs) -> websocket.WebSocket:
        """
        Create a websocket that uses the session credentials to establish
        itself.

        :param str url: Actual location to open websocket at
        :param dict(str,str) header: Optional HTTP headers
        :param str cookie:
            Optional cookies (composed as semicolon-separated string)
        :param kwargs: Other options to :py:func:`~websocket.create_connection`
        :rtype: ~websocket.WebSocket
        """
        # Note: *NOT* a renewable action!
        if header is None:
            header = {}
        header[self.__csrf_header] = self.__csrf
        if cookie is not None:
            cookie += ";" + _SESSION_COOKIE + "=" + self._session_id
        else:
            cookie = _SESSION_COOKIE + "=" + self._session_id
        return websocket.create_connection(
            url, header=header, cookie=cookie, **kwargs)

    def _purge(self):
        """
        Clears out all credentials from this session, rendering the session
        completely inoperable henceforth.
        """
        self.__username = None
        self.__password = None
        self._session_id = None
        self.__csrf = None


class SessionAware:
    """
    Connects to the session.

    .. warning::
        This class does not present a stable API for public consumption.
    """
    __slots__ = ("__session", "_url")

    def __init__(self, session: Session, url: str):
        self.__session = session
        self._url = clean_url(url)

    @property
    def _session_credentials(self):
        """
        The current session credentials.
        Only supposed to be called by subclasses.

        :rtype: tuple(dict(str,str),dict(str,str))
        """
        # pylint: disable=protected-access
        return self.__session._credentials

    @property
    def _service_url(self):
        """
        The main service URL.

        :rtype: str
        """
        # pylint: disable=protected-access
        return self.__session._service_url

    def _get(self, url: str, **kwargs) -> requests.Response:
        return self.__session.get(url, **kwargs)

    def _post(self, url: str, json_dict: dict, **kwargs) -> requests.Response:
        return self.__session.post(url, json_dict, **kwargs)

    def _put(self, url: str, data: str, **kwargs) -> requests.Response:
        return self.__session.put(url, data, **kwargs)

    def _delete(self, url: str, **kwargs) -> requests.Response:
        return self.__session.delete(url, **kwargs)

    def _websocket(self, url: str, **kwargs) -> websocket.WebSocket:
        """
        Create a websocket that uses the session credentials to establish
        itself.

        :param str url: Actual location to open websocket at
        :rtype: ~websocket.WebSocket
        """
        return self.__session.websocket(url, **kwargs)
